import pytest

import sdmx
from sdmx.message import Message


@pytest.mark.parametrize_specimens("path", format="xml")
def test_read_xml(path) -> None:
    """XML specimens can be read."""
    if "esms_structured" in path.name or "query" in str(path):
        pytest.xfail("Not implemented")

    result = sdmx.read_sdmx(path)
    assert isinstance(result, Message)


@pytest.mark.parametrize(
    "message_path, structure_path, structure_id",
    (
        # Structure-specific data; same as test_reader_xml_v21.test_read_ss_xml but
        # without additional assertions
        ("M.USD.EUR.SP00.A.xml", "ECB_EXR/1/structure.xml", "ECB_EXR1"),
        # Structure-specific metadata
        ("esms_structured.xml", "v21/xml/demography/esms.xml", "ESMS_SIMPLE"),
    ),
)
def test_read_xml_ss(specimen, message_path, structure_path, structure_id) -> None:
    """Structure-specific (meta)data specimens can be read *using* their structures.

    Note that :func:`.test_read_xml` tests some of the same {Metad,D}ataMessage
    specimens, but *without* supplying the [MD]SD; in those cases, the [MD]SD contents
    are inferred while parsing. This test, in contrast, parses with the [MD]SD
    available.
    """
    # Read the structure message and retrieve the structure object
    with specimen(structure_path) as f:
        sm = sdmx.read_sdmx(f)

    # Structure may be for data or metadata
    for name in "structure", "metadatastructure":
        try:
            s = getattr(sm, name)[structure_id]
        except KeyError:
            pass
        else:
            break

    assert s

    # The (meta)data message can be read using its associated structure
    with specimen(message_path) as f:
        result = sdmx.read_sdmx(f, dsd=s)

    assert isinstance(result, Message)



"""Parse structure information from item description."""
import logging
import operator
import re
from typing import Callable, Dict, List, Optional, Tuple, Union

from sdmx.model.common import DEFAULT_LOCALE, Item, ItemScheme

log = logging.getLogger(__name__)


#: Operators understood by :func:`parse_item_description`
OPS = {"+": operator.add, "-": operator.sub, "=": operator.eq}


def parse_item_description(
    item: Item, locale: Optional[str] = None
) -> List[Tuple[Callable, str]]:
    """Parse the :attr:`.description` of `item` for a structure expression.

    A common—but **non-standard**—SDMX usage is that :class:`Items <.Item>` in
    :class:`ItemSchemes <.ItemScheme>` contain structure expressions in their
    :attr:`description <.NameableArtefact.description>`. These may resemble::

        A = B + C - D

    …indicating that data for ``A`` can be computed by adding data for ``B`` and ``C``
    and subtracting data for ``D``. In this usage, ``A`` matches the
    :attr:`.IdentifiableArtefact.id` of the `item`, and ``B``, ``C``, and ``D`` are
    usually the IDs of other Items in the same ItemScheme.

    Another form is::

        B + C - D

    In this case, the left-hand side and "=" are omitted.

    :func:`parse_item_description` parses these expressions, returning a list of tuples.
    In each tuple, the first element gives the operation to be applied (in the above,
    examples, implicitly :func:`~operator.add` for "B"), and the second element is the
    ID of the operand.

    Other descriptions are not (yet) supported, including:

    - Multi-line descriptions, in which the structure expression occurs on one line.
    - Structure expressions that appear on lines with other text, for example::

          Some text; A = B + C - D

    Parameters
    ----------
    locale : str
        Use this :attr:`localization <.localizations>` of `item`'s description, instead
        of the default (:obj:`.DEFAULT_LOCALE` or the sole localization).

    Returns
    -------
    list of tuple of (operator, str)
        The list is empty if:

        - `item` has no :attr:`description`.
        - The description does not contain a structure expression.
        - An expression like "A = B + …" exists, but ``A`` does not match the ID of
          `item`.

    Example
    -------
    >>> from sdmx.model.common import Code
    >>> from sdmx.util.item_structure import parse_item_description
    >>> c = Code(id="A", description="A = B + C - D")
    >>> parse_item_description(c)
    [(<function _operator.add(a, b, /)>, 'B'),
     (<function _operator.add(a, b, /)>, 'C'),
     (<function _operator.sub(a, b, /)>, 'D')]

    """
    desc = item.description

    if len(desc.localizations) == 0:
        return []
    elif locale:
        text = desc.localizations[locale]
    else:
        if len(desc.localizations) > 1 and not locale:
            log.warning(
                f">1 localization for {item}.description; using "
                f".localized_default({DEFAULT_LOCALE!r})"
            )
        text = str(desc)

    # Split the text by operator-like expressions (+, -, =) and handle tokens
    result = []
    current_op = operator.add
    current_id = ""
    for token in re.split(r"\s*([\+=-])\s*", text):
        op = OPS.get(token)
        if op is None:
            # The ID of another Item in the ItemScheme → store
            current_id = token
        elif op is operator.eq:
            # An "=" separating the LHS and RHS of the equation → check and discard
            if current_id != item.id:
                log.debug(
                    f'Left-hand side "{current_id} = " does not match ID of {item!r}'
                )
                return []
        else:
            # Any other operator → add the previous (operator, ID) to result; store
            result.append((current_op, current_id))
            current_op = op

    # Add remaining item, if any
    if current_id:
        result.append((current_op, current_id))

    return result


def parse_item(
    itemscheme: ItemScheme, id=str, **kwargs
) -> List[Tuple[Callable, Union[Item, str]]]:
    """Parse a structure expression for the item in `itemscheme` with the given `id`.

    In addition to the behaviour of :func:`parse_item_description`, :func:`parse_item`
    replaces—where possible—the operand :class:`str` IDs with references to specific
    other :class:`Items <.Item>` in the `itemscheme`.

    Where not possible, the operand is returned as-is. For example, in an expression
    like::

        A = B + C - D (until 2022) - E

    …``B``, ``C``, and ``E`` may resolve to references to particular other items, while
    the string "D (until 2022)" will be returned as-is, and not further parsed as a
    reference to an item with ID ``D``. (This functionality may be added in a future
    version of :mod:`sdmx`.)

    Parameters
    ----------
    kwargs :
        Passed to :func:`parse_item_description`.

    Returns
    -------
    list of tuple of (operator, operand)
        Where possible, `operand` is a reference to an existing item in `itemscheme`;
        otherwise the same :class:`str` as returned by :func:`parse_item_description`.
    """

    # Validate
    result, missing = [], []
    for op, operand in parse_item_description(itemscheme[id], **kwargs):
        try:
            # Retrieve the item with ID matching `operand`
            item = itemscheme[operand]
        except KeyError:
            # No such item
            missing.append(operand)
            item = operand
        # Store
        result.append((op, item))

    if len(missing) == len(result) > 0:
        log.debug(f'No tokens matching IDs in "{itemscheme[id].description!s}"')
        return []
    elif len(missing):
        log.debug(
            f"In expression for {itemscheme}[{id}], could not parse item(s): "
            + ", ".join(map(repr, missing))
        )

    return result


def parse_all(
    itemscheme: ItemScheme, **kwargs
) -> Dict[str, List[Tuple[Callable, Union[Item, str]]]]:
    """Parse structure expressions for every item in `itemscheme`.

    Parameters
    ----------
    kwargs :
        Passed to :func:`parse_item_description` via :func:`parse_item`.

    Returns
    -------
    dict of (Item → list)
        Keys are references to the items of `itemscheme`. Values are the results of
        :func:`parse_item`, i.e. possibly empty.
    """
    return {item: parse_item(itemscheme, item.id) for item in itemscheme}



"""SDMX 2.1 Information Model."""
import logging

# TODO for complete implementation of the IM, enforce TimeKeyValue (instead of KeyValue)
#      for {Generic,StructureSpecific} TimeSeriesDataSet.
from dataclasses import dataclass, field
from typing import (
    ClassVar,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

from sdmx.dictlike import DictLikeDescriptor

from . import common
from .common import (
    IT,
    AttributeRelationship,
    Component,
    ComponentList,
    ConstrainableArtefact,
    ConstraintRole,
    DataAttribute,
    DimensionComponent,
    IdentifiableArtefact,
    Key,
    NameableArtefact,
)

# Classes defined directly in the current file, in the order they appear
__all__ = [
    "SelectionValue",
    "MemberValue",
    "TimeRangeValue",
    "BeforePeriod",
    "AfterPeriod",
    "RangePeriod",
    "DataKey",
    "DataKeySet",
    "Constraint",
    "MemberSelection",
    "ContentConstraint",
    "MeasureDimension",
    "PrimaryMeasure",
    "MeasureDescriptor",
    "NoSpecifiedRelationship",
    "PrimaryMeasureRelationship",
    "ReportingYearStartDay",
    "DataStructureDefinition",
    "DataflowDefinition",
    "Observation",
    "StructureSpecificDataSet",
    "GenericDataSet",
    "GenericTimeSeriesDataSet",
    "StructureSpecificTimeSeriesDataSet",
    "ReportingCategory",
    "ReportingTaxonomy",
    "TargetObject",
    "DataSetTarget",
    "DimensionDescriptorValuesTarget",
    "IdentifiableObjectTarget",
    "ReportPeriodTarget",
    "MetadataTarget",
    "ReportStructure",
    "MetadataStructureDefinition",
    "MetadataflowDefinition",
    "TargetObjectValue",
    "TargetReportPeriod",
    "TargetIdentifiableObject",
    "TargetObjectKey",
    "ReportedAttribute",
    "EnumeratedAttributeValue",
    "NonEnumeratedAttributeValue",
    "OtherNonEnumeratedAttributeValue",
    "TextAttributeValue",
    "XHTMLAttributeValue",
    "MetadataReport",
    "MetadataSet",
    "Hierarchy",
    "HierarchicalCodelist",
    "ItemAssociation",
    "CodeMap",
    "ItemSchemeMap",
    "CodelistMap",
    "StructureSet",
]

log = logging.getLogger(__name__)


# §10.3: Constraints


class SelectionValue(common.BaseSelectionValue):
    """SDMX 2.1 SelectionValue.

    Identical to its parent class.
    """


class MemberValue(common.BaseMemberValue, SelectionValue):
    """SDMX 2.1 MemberValue."""


class TimeRangeValue(SelectionValue):
    """SDMX 2.1 TimeRangeValue."""


class BeforePeriod(TimeRangeValue, common.Period):
    pass


class AfterPeriod(TimeRangeValue, common.Period):
    pass


@dataclass
class RangePeriod(TimeRangeValue):
    start: common.StartPeriod
    end: common.EndPeriod


class DataKey(common.BaseDataKey):
    """SDMX 2.1 DataKey.

    Identical to its parent class.
    """


class DataKeySet(common.BaseDataKeySet):
    """SDMX 2.1 DataKeySet.

    Identical to its parent class.
    """


@dataclass
class Constraint(common.BaseConstraint):
    """SDMX 2.1 Constraint.

    For SDMX 3.0, see :class:`.v30.Constraint`.
    """

    # NB the spec gives 1..* for this attribute, but this implementation allows only 1
    role: Optional[ConstraintRole] = None
    #: :class:`.DataKeySet` included in the Constraint.
    data_content_keys: Optional[DataKeySet] = None
    # metadata_content_keys: MetadataKeySet = None

    def __contains__(self, value):
        if self.data_content_keys is None:
            raise NotImplementedError("Constraint does not contain a DataKeySet")

        return value in self.data_content_keys


class MemberSelection(common.BaseMemberSelection):
    """SDMX 2.1 MemberSelection."""


@dataclass
@NameableArtefact._preserve("repr")
class ContentConstraint(Constraint, common.BaseContentConstraint):
    #: :class:`CubeRegions <.CubeRegion>` included in the ContentConstraint.
    data_content_region: List[common.CubeRegion] = field(default_factory=list)
    #:
    content: Set[ConstrainableArtefact] = field(default_factory=set)
    metadata_content_region: Optional[common.MetadataTargetRegion] = None

    def __contains__(self, value):
        if self.data_content_region:
            return all(value in cr for cr in self.data_content_region)
        else:
            raise NotImplementedError("ContentConstraint does not contain a CubeRegion")

    def to_query_string(self, structure):
        cr_count = len(self.data_content_region)
        try:
            if cr_count > 1:
                log.warning(f"to_query_string() using first of {cr_count} CubeRegions")

            return self.data_content_region[0].to_query_string(structure)
        except IndexError:
            raise RuntimeError("ContentConstraint does not contain a CubeRegion")

    def iter_keys(
        self,
        obj: Union["DataStructureDefinition", "DataflowDefinition"],
        dims: List[str] = [],
    ) -> Generator[Key, None, None]:
        """Iterate over keys.

        A warning is logged if `obj` is not already explicitly associated to this
        ContentConstraint, i.e. present in :attr:`.content`.

        See also
        --------
        .DataStructureDefinition.iter_keys
        """
        if obj not in self.content:
            log.warning(f"{repr(obj)} is not in {repr(self)}.content")

        yield from obj.iter_keys(constraint=self, dims=dims)


# §5.3: Data Structure Definition


@dataclass
class MeasureDimension(DimensionComponent):
    """SDMX 2.1 MeasureDimension.

    This class is not present in SDMX 3.0.
    """

    #:
    concept_role: Optional[common.Concept] = None


class PrimaryMeasure(Component):
    """SDMX 2.1 PrimaryMeasure.

    This class is not present in SDMX 3.0; see instead :class:`.v30.Measure`.
    """


class MeasureDescriptor(ComponentList[PrimaryMeasure]):
    """SDMX 2.1 MeasureDescriptor.

    For SDMX 3.0 see instead :class:`.v30.MeasureDescriptor`.
    """

    _Component = PrimaryMeasure


class NoSpecifiedRelationship(AttributeRelationship):
    """Indicates that the attribute is attached to the entire data set."""


class PrimaryMeasureRelationship(AttributeRelationship):
    """Indicates that the attribute is attached to a particular observation."""


class ReportingYearStartDay(DataAttribute):
    """SDMX 2.1 ReportingYearStartDay.

    This class is deleted in SDMX 3.0.
    """


@dataclass(repr=False)
@IdentifiableArtefact._preserve("hash")
class DataStructureDefinition(common.BaseDataStructureDefinition):
    """SDMX 2.1 DataStructureDefinition (‘DSD’)."""

    MemberValue = MemberValue
    MemberSelection = MemberSelection
    ConstraintType = ContentConstraint

    #: A :class:`.MeasureDescriptor`.
    measures: MeasureDescriptor = field(default_factory=MeasureDescriptor)


@dataclass(repr=False)
@IdentifiableArtefact._preserve("hash")
class DataflowDefinition(common.BaseDataflow):
    #:
    structure: DataStructureDefinition = field(default_factory=DataStructureDefinition)


# §5.4: Data Set


@dataclass
class Observation(common.BaseObservation):
    #:
    value_for: Optional[PrimaryMeasure] = None


@dataclass
class DataSet(common.BaseDataSet):
    """SDMX 2.1 DataSet."""

    #: Named ``attachedAttribute`` in the IM.
    attrib: DictLikeDescriptor[str, common.AttributeValue] = DictLikeDescriptor()


class StructureSpecificDataSet(DataSet):
    """SDMX 2.1 StructureSpecificDataSet.

    This subclass has no additional functionality compared to DataSet.
    """


class GenericDataSet(DataSet):
    """SDMX 2.1 GenericDataSet.

    This subclass has no additional functionality compared to DataSet.
    """


class GenericTimeSeriesDataSet(DataSet):
    """SDMX 2.1 GenericTimeSeriesDataSet.

    This subclass has no additional functionality compared to DataSet.
    """


class StructureSpecificTimeSeriesDataSet(DataSet):
    """SDMX 2.1 StructureSpecificTimeSeriesDataSet.

    This subclass has no additional functionality compared to DataSet.
    """


# §7.3 Metadata Structure Definition


class ReportingCategory(common.Item):
    """SDMX 2.1 ReportingCategory."""


class ReportingTaxonomy(common.ItemScheme):
    """SDMX 2.1 ReportingTaxonomy."""


class TargetObject(common.Component):
    """SDMX 2.1 TargetObject."""


class DataSetTarget(TargetObject):
    """SDMX 2.1 DataSetTarget."""


class DimensionDescriptorValuesTarget(TargetObject):
    """SDMX 2.1 DimensionDescriptorValuesTarget."""


class IdentifiableObjectTarget(TargetObject):
    """SDMX 2.1 IdentifiableObjectTarget."""


class ReportPeriodTarget(TargetObject):
    """SDMX 2.1 ReportPeriodTarget."""


class MetadataTarget(ComponentList):
    """SDMX 2.1 MetadataTarget."""

    _Component = TargetObject


@dataclass
class ReportStructure(ComponentList):
    """SDMX 2.1 ReportStructure."""

    _Component = common.MetadataAttribute

    report_for: List[MetadataTarget] = field(default_factory=list)


@dataclass
@common.MaintainableArtefact._preserve("hash")
class MetadataStructureDefinition(common.BaseMetadataStructureDefinition):
    """SDMX 2.1 MetadataStructureDefinition."""

    report_structure: DictLikeDescriptor[str, ReportStructure] = DictLikeDescriptor()

    #: Association to 1 or more :class:`.MetadataTarget`
    target: DictLikeDescriptor[str, MetadataTarget] = DictLikeDescriptor()


@dataclass
class MetadataflowDefinition(common.BaseMetadataflow):
    """SDMX 2.1 MetadataflowDefinition."""

    # NB narrows the type of common.StructureUsage.structure
    structure: Optional[MetadataStructureDefinition] = None


# §7.4: Metadata Set


@dataclass
class TargetObjectValue:
    """SDMX 2.1 TargetObjectValue."""

    value_for: TargetObject


@dataclass
class TargetReportPeriod(TargetObjectValue):
    """SDMX 2.1 TargetReportPeriod."""

    report_period: str


@dataclass
class TargetIdentifiableObject(TargetObjectValue):
    """SDMX 2.1 TargetIdentifiableObject."""

    obj: IdentifiableArtefact


@dataclass
class TargetObjectKey:
    """SDMX 2.1 TargetObjectKey."""

    key_values: DictLikeDescriptor[str, TargetObjectValue] = DictLikeDescriptor()


@dataclass
class ReportedAttribute:
    """SDMX 2.1 ReportedAttribute.

    Analogous to :class:`.v30.MetadataAttributeValue`.
    """

    value_for: common.MetadataAttribute
    parent: Optional["ReportedAttribute"] = None
    child: List["ReportedAttribute"] = field(default_factory=list)

    def __getitem__(self, index: int) -> "ReportedAttribute":
        return self.child[index]

    def __len__(self) -> int:
        return len(self.child)


class EnumeratedAttributeValue(ReportedAttribute):
    """SDMX 2.1 EnumeratedAttributeValue.

    Analogous to :class:`.v30.CodedMetadataAttributeValue`.
    """

    value: str

    #: .. note::
    #:
    #:    The SDMX 2.1 IM (2011-08) gives this as `valueFor`, but this name duplicates
    #:    :attr:`.ReportedAttribute.value_for`. :mod:`sdmx` uses `value_of` for
    #:    consistency with :attr:`.v30.CodedMetadataAttributeValue.value_of`.
    value_of: common.Code


class NonEnumeratedAttributeValue(ReportedAttribute):
    """SDMX 2.1 NonEnumeratedAttributeValue."""


class OtherNonEnumeratedAttributeValue(NonEnumeratedAttributeValue):
    """SDMX 2.1 OtherNonEnumeratedAttributeValue."""

    value: str


class TextAttributeValue(NonEnumeratedAttributeValue, common.BaseTextAttributeValue):
    """SDMX 2.1 TextAttributeValue."""


@dataclass
class XHTMLAttributeValue(NonEnumeratedAttributeValue, common.BaseXHTMLAttributeValue):
    """SDMX 2.1 XHTMLAttributeValue."""

    value: str


@dataclass
class MetadataReport:
    """SDMX 2.1 MetadataReport."""

    metadata: List[ReportedAttribute] = field(default_factory=list)
    target: Optional[MetadataTarget] = None
    attaches_to: Optional[TargetObjectKey] = None


@dataclass
class MetadataSet(NameableArtefact, common.BaseMetadataSet):
    """SDMX 2.1 MetadataSet.

    .. note:: Contrast :class:`.v30.MetadataSet`, which is a
       :class:`.MaintainableArtefact`.
    """

    described_by: Optional[MetadataflowDefinition] = None
    # described_by: Optional[ReportStructure] = None
    structured_by: Optional[MetadataStructureDefinition] = None

    #: Analogous to :attr:`.v30.MetadataSet.provided_by`.
    published_by: Optional[common.DataProvider] = None

    report: List[MetadataReport] = field(default_factory=list)


# §8 Hierarchical Code List


@dataclass
class Hierarchy(NameableArtefact):
    """SDMX 2.1 Hierarchy."""

    has_formal_levels: bool = False

    #: Hierarchical codes in the hierarchy.
    codes: Dict[str, common.HierarchicalCode] = field(default_factory=dict)

    level: Optional[common.Level] = None


@dataclass
class HierarchicalCodelist(common.MaintainableArtefact):
    """SDMX 2.1 HierarchicalCodelist."""

    hierarchy: List[Hierarchy] = field(default_factory=list)

    def __repr__(self) -> str:
        tmp = super(NameableArtefact, self).__repr__()[:-1]
        return f"{tmp}: {len(self.hierarchy)} hierarchies>"


# §9: Structure Set and Mappings


@dataclass
class ItemAssociation(common.AnnotableArtefact, Generic[IT]):
    """SDMX 2.1 ItemAssociation."""

    _Item: ClassVar[Type[common.Item]] = common.Item

    source: Optional[IT] = None
    target: Optional[IT] = None


class CodeMap(ItemAssociation[common.Code]):
    """SDMX 2.1 CodeMap."""

    _Item = common.Code


IAT = TypeVar("IAT", bound="ItemAssociation")
IST = TypeVar("IST", bound="common.ItemScheme")


@dataclass
class ItemSchemeMap(NameableArtefact, Generic[IST, IAT]):
    """SDMX 2.1 ItemSchemeMap."""

    _ItemAssociation: ClassVar[Type[ItemAssociation]] = ItemAssociation

    source: Optional[IST] = None
    target: Optional[IST] = None

    item_association: List[IAT] = field(default_factory=list)


class CodelistMap(ItemSchemeMap[common.Codelist, CodeMap]):
    """SDMX 2.1 CodelistMap."""

    _ItemAssociation = CodeMap


@dataclass
class StructureSet(common.MaintainableArtefact):
    """SDMX 2.1 StructureSet."""

    item_scheme_map: List[ItemSchemeMap] = field(default_factory=list)


CF = common.ClassFinder(
    __name__,
    name_map={
        "Dataflow": "DataflowDefinition",
        "DataStructure": "DataStructureDefinition",
        "MetadataStructure": "MetadataStructureDefinition",
        "Metadataflow": "MetadataflowDefinition",
    },
    parent_map={
        common.HierarchicalCode: Hierarchy,
        PrimaryMeasure: MeasureDescriptor,
        MetadataTarget: MetadataStructureDefinition,
    },
)
get_class = CF.get_class
parent_class = CF.parent_class
__dir__ = CF.dir
__getattr__ = CF.getattr



import io
import re
from pathlib import Path

import pytest

import sdmx
from sdmx.format import Version, xml
from sdmx.format.xml.common import _extracted_zipball
from sdmx.message import StructureMessage
from sdmx.model import v21


def test_ns_prefix():
    with pytest.raises(ValueError):
        xml.v21.ns_prefix("https://example.com")


def test_qname():
    assert f"{{{xml.v21.base_ns}/structure}}Code" == str(xml.v21.qname("str", "Code"))
    assert f"{{{xml.v30.base_ns}/structure}}Code" == str(xml.v30.qname("str", "Code"))


def test_tag_for_class():
    # ItemScheme is never written to XML; no corresponding tag name
    assert xml.v21.tag_for_class(v21.ItemScheme) is None


def test_class_for_tag():
    assert xml.v30.class_for_tag("str:DataStructure") is not None


@pytest.fixture(scope="module")
def mock_gh_api():
    """Mock GitHub API responses to avoid hitting rate limits.

    For each API endpoint URL queried by :func:.`_gh_zipball`, return a pared-down JSON
    response that contains the required "zipball_url" key.
    """
    import requests_mock

    base = "https://api.github.com/repos/sdmx-twg/sdmx-ml"

    with requests_mock.Mocker(real_http=True) as m:
        for v in "2.1", "3.0":
            m.get(
                f"{base}/releases/tags/v{v}",
                json=dict(zipball_url=f"{base}/zipball/v{v}"),
            )
        yield


@pytest.fixture(scope="module")
def installed_schemas(mock_gh_api, tmp_path_factory):
    """Fixture that ensures schemas are installed locally in a temporary directory."""
    dir = tmp_path_factory.mktemp("schemas")
    sdmx.install_schemas(dir.joinpath("2.1"), Version["2.1"])
    sdmx.install_schemas(dir.joinpath("3.0"), Version["3.0.0"])
    yield dir


@pytest.mark.parametrize("version", ["1", 1, None])
def test_install_schemas_invalid_version(version):
    """Ensure invalid versions throw ``NotImplementedError``."""
    with pytest.raises(NotImplementedError):
        sdmx.install_schemas(version=version)


@pytest.mark.network
@pytest.mark.parametrize("version", ["2.1", "3.0"])
def test_install_schemas(installed_schemas, version):
    """Test that XSD files are downloaded and ready for use in validation."""
    # Look for a couple of the expected files
    for schema_doc in ("SDMXCommon.xsd", "SDMXMessage.xsd"):
        assert installed_schemas.joinpath(version, schema_doc).exists()


@pytest.mark.network
def test_install_schemas_in_user_cache():
    """Test that XSD files are downloaded and ready for use in validation."""
    import platformdirs

    cache_dir = platformdirs.user_cache_path("sdmx") / "2.1"
    sdmx.install_schemas()

    # Look for a couple of the expected files
    files = ["SDMXCommon.xsd", "SDMXMessage.xsd"]
    for schema_doc in files:
        doc = cache_dir.joinpath(schema_doc)
        assert doc.exists(), (cache_dir, sorted(cache_dir.glob("*")))


@pytest.mark.parametrize("version", ["1", 1, None])
def test_validate_xml_invalid_version(version):
    """Ensure validation of invalid versions throw ``NotImplementedError``."""
    with pytest.raises(NotImplementedError):
        # This message doesn't exist, but the version should throw before it is used.
        sdmx.validate_xml("samples/common/common.xml", version=version)


def test_validate_xml_no_schemas(tmp_path, specimen, installed_schemas):
    """Check that supplying an invalid schema path will raise ``ValueError``."""
    with specimen("IPI-2010-A21-structure.xml", opened=False) as msg_path:
        with pytest.raises(ValueError):
            sdmx.validate_xml(msg_path, schema_dir=tmp_path)


@pytest.mark.network
def test_validate_xml_from_v2_1_samples(tmp_path, specimen, installed_schemas):
    """Use official samples to ensure validation of v2.1 messages works correctly."""
    extracted_content = _extracted_zipball(Version["2.1"])

    # Schemas as just in a flat directory
    schema_dir = extracted_content.joinpath("schemas")

    # Samples are somewhat spread out, and some are known broken so we pick a bunch
    for parts in [
        ("v21", "xml", "common", "common.xml"),
        ("v21", "xml", "demography", "demography.xml"),
        ("v21", "xml", "demography", "esms.xml"),
        ("ECB_EXR", "common.xml"),
        ("ECB_EXR", "ng-structure-full.xml"),
        ("ECB_EXR", "ng-structure.xml"),
        ("v21", "xml", "query", "query_cl_all.xml"),
        ("v21", "xml", "query", "response_cl_all.xml"),
        ("v21", "xml", "query", "query_esms_children.xml"),
        ("v21", "xml", "query", "response_esms_children.xml"),
    ]:
        with specimen(str(Path(*parts))) as sample:
            assert sdmx.validate_xml(sample, schema_dir, version="2.1")


@pytest.mark.network
def test_validate_xml_invalid_doc(tmp_path, installed_schemas):
    """Ensure that an invalid document fails validation."""
    msg_path = tmp_path / "invalid.xml"

    # Generate a codelist to form a StructureMessage
    ECB = v21.Agency(id="ECB")

    cl = v21.Codelist(
        id="CL_COLLECTION",
        version="1.0",
        is_final=False,
        is_external_reference=False,
        maintainer=ECB,
        name={"en": "Collection indicator code list"},
    )

    # Add items
    CL_ITEMS = [
        dict(id="A", name={"en": "Average of observations through period"}),
        dict(id="B", name={"en": "Beginning of period"}),
        dict(id="B1", name={"en": "Child code of B"}),
    ]
    for info in CL_ITEMS:
        cl.items[info["id"]] = v21.Code(**info)

    msg = StructureMessage(codelist={cl.id: cl})

    msg_path.write_bytes(sdmx.to_xml(msg))

    # Expect validation to fail
    assert not sdmx.validate_xml(msg_path, schema_dir=installed_schemas.joinpath("2.1"))


def test_validate_xml_invalid_message_type():
    """Ensure that an invalid document fails validation."""
    # Create a mangled structure message with its outmost tag changed to be invalid
    msg = StructureMessage()
    invalid_msg = re.sub(b"mes:Structure([ >])", rb"mes:FooBar\1", sdmx.to_xml(msg))

    with pytest.raises(NotImplementedError):
        sdmx.validate_xml(io.BytesIO(invalid_msg))


@pytest.mark.network
def test_validate_xml_from_v3_0_samples(tmp_path, installed_schemas):
    """Use official samples to ensure validation of v3.0 messages works correctly."""
    extracted_content = _extracted_zipball(Version["3.0.0"])

    # Schemas as just in a flat directory
    schema_dir = extracted_content.joinpath("schemas")

    # Samples are somewhat spread out, and some are known broken so we pick a bunch
    samples_dir = extracted_content.joinpath("samples")
    samples = [
        samples_dir / "Codelist" / "codelist.xml",
        samples_dir / "Codelist" / "codelist - extended.xml",
        samples_dir / "Concept Scheme" / "conceptscheme.xml",
        samples_dir / "Data Structure Definition" / "ECB_EXR.xml",
        samples_dir / "Dataflow" / "dataflow.xml",
        samples_dir / "Geospatial" / "geospatial_geographiccodelist.xml",
    ]
    for sample in samples:
        assert sdmx.validate_xml(sample, schema_dir, version="3.0")



from sdmx.format import list_media_types
from sdmx.reader.base import BaseReader


class Reader(BaseReader):
    """Stub (incomplete) implementation of a SDMX-CSV reader."""

    media_types = list_media_types(base="csv")
    suffixes = [".csv"]

    def read_message(self, source, dsd=None):  # pragma: no cover
        """Not implemented."""
        raise NotImplementedError



import logging
import re
import zipfile
from functools import lru_cache
from itertools import chain
from operator import itemgetter
from pathlib import Path
from shutil import copytree
from typing import IO, Iterable, List, Mapping, Optional, Tuple, Union

from lxml import etree
from lxml.etree import QName

from sdmx.format import Version

log = logging.getLogger(__name__)

# Tags common to SDMX-ML 2.1 and 3.0

# XML tag name ("str" namespace) and class name are the same
CT1 = [
    "Agency",
    "AgencyScheme",
    "Categorisation",
    "Category",
    "CategoryScheme",
    "Code",
    "Codelist",
    "Concept",
    "ConceptScheme",
    "CustomType",
    "CustomTypeScheme",
    "DataConsumer",
    "DataConsumerScheme",
    "DataProvider",
    "DataProviderScheme",
    "HierarchicalCode",
    "Level",
    "NamePersonalisation",
    "NamePersonalisationScheme",
    "Ruleset",
    "RulesetScheme",
    "TimeDimension",
    "TransformationScheme",
    "UserDefinedOperatorScheme",
]

# XML tag name and class name differ
CT2 = [
    ("model.Agency", "str:Agency"),  # Order matters
    ("model.Agency", "mes:Receiver"),
    ("model.Agency", "mes:Sender"),
    ("model.AttributeDescriptor", "str:AttributeList"),
    ("model.Concept", "str:ConceptIdentity"),
    ("model.Codelist", "str:Enumeration"),  # This could possibly be ItemScheme
    ("model.Dimension", "str:Dimension"),  # Order matters
    ("model.Dimension", "str:DimensionReference"),
    ("model.Dimension", "str:GroupDimension"),
    ("model.DataAttribute", "str:Attribute"),
    ("model.DataStructureDefinition", "str:DataStructure"),
    ("model.DimensionDescriptor", "str:DimensionList"),
    ("model.GroupDimensionDescriptor", "str:Group"),
    ("model.GroupDimensionDescriptor", "str:AttachmentGroup"),
    ("model.GroupKey", "gen:GroupKey"),
    ("model.Key", "gen:ObsKey"),
    ("model.MeasureDescriptor", "str:MeasureList"),
    ("model.MetadataStructureDefinition", "str:MetadataStructure"),
    ("model.SeriesKey", "gen:SeriesKey"),
    ("model.Structure", "com:Structure"),
    ("model.Structure", "str:Structure"),
    ("model.StructureUsage", "com:StructureUsage"),
    ("model.VTLMappingScheme", "str:VtlMappingScheme"),
    # Message classes
    ("message.DataMessage", "mes:StructureSpecificData"),
    ("message.MetadataMessage", "mes:GenericMetadata"),
    ("message.MetadataMessage", "mes:StructureSpecificMetadata"),
    ("message.ErrorMessage", "mes:Error"),
    ("message.StructureMessage", "mes:Structure"),
]

NS = {
    "": None,
    "xml": "http://www.w3.org/XML/1998/namespace",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    # To be formatted
    "com": "{}/common",
    "md": "{}/metadata/generic",
    "data": "{}/data/structurespecific",
    "str": "{}/structure",
    "mes": "{}/message",
    "gen": "{}/data/generic",
    "footer": "{}/message/footer",
}


def validate_xml(
    msg: Union[Path, IO],
    schema_dir: Optional[Path] = None,
    version: Union[str, Version] = Version["2.1"],
) -> bool:
    """Validate and SDMX message against the XML Schema (XSD) documents.

    The XML Schemas must first be installed or validation will fail. See
    :func:`sdmx.install_schemas` to download the schema files.

    Parameters
    ----------
    msg
        A SDMX-ML Message formatted XML file.
    schema_dir
        The directory to XSD schemas used to validate the message.
    version
        The SDMX-ML schema version to validate against. One of ``2.1`` or ``3.0``.

    Returns
    -------
    bool
        True if validation passed. False otherwise.
    """
    schema_dir, version = _handle_validate_args(schema_dir, version)

    msg_doc = etree.parse(msg)

    # Make sure the message is a supported type
    supported_elements = [
        "CodelistQuery",
        "DataStructureQuery",
        "GenericData",
        "GenericMetadata",
        "GenericTimeSeriesData",
        "MetadataStructureQuery",
        "Structure",
        "StructureSpecificData",
        "StructureSpecificMetadata",
        "StructureSpecificTimeSeriesData",
    ]
    root_elem_name = msg_doc.docinfo.root_name
    if root_elem_name not in supported_elements:
        raise NotImplementedError

    message_xsd = schema_dir.joinpath("SDMXMessage.xsd")
    if not message_xsd.exists():
        raise ValueError(f"Could not find XSD files in {schema_dir}")

    # Turn the XSD into a schema object
    xml_schema_doc = etree.parse(message_xsd)
    xml_schema = etree.XMLSchema(xml_schema_doc)

    try:
        xml_schema.assertValid(msg_doc)
    except etree.DocumentInvalid as err:
        log.error(err)
    finally:
        return xml_schema.validate(msg_doc)


def _extracted_zipball(version: Version) -> Path:
    """Retrieve, cache, and extract the SDMX-ML schemas for `version`.

    1. Query the GitHub REST API to identify a URL for the `version` in zipball format.
    2. Download and cache the zipball. The file is not downloaded if it already exists.
    3. Unpack the archive.

    Actions (2) and (3) are performed in the user's cache directory (for instance,
    :file:`$HOME/.cache/sdmx/`). :func:`install_schemas` handles copying the extracted
    files to other locations.

    Returns
    -------
    Path
        Path to the root folder of the unpacked archive.
    """
    import platformdirs
    import requests

    # Map SDMX-ML schema versions to repo paths
    # Check the latest release to get the URL to the schema zip
    url = (
        "https://api.github.com/repos/sdmx-twg/sdmx-ml/releases/tags/"
        + {Version["2.1"]: "v2.1", Version["3.0.0"]: "v3.0.0"}[version]
    )
    gh_headers = {
        "Accept": "application/vnd.github+json",
        "X-GitHub-Api-Version": "2022-11-28",
    }
    release_json = requests.get(url=url, headers=gh_headers).json()
    try:
        zipball_url = release_json["zipball_url"]
    except KeyError:  # pragma: no cover
        log.debug(release_json)
        raise RuntimeError("Failed to download SDMX-ML schema bundle")

    # Make a request for the zipball
    resp = requests.get(url=zipball_url, headers=gh_headers)

    # Filename indicated by the HTTP response
    filename = resp.headers["content-disposition"].split("filename=")[-1]
    # Location for the cached zipball
    target = platformdirs.user_cache_path("sdmx").joinpath(filename)

    # Avoid downloading if the same file is already present
    if target.exists():
        log.info(f"Use existing {target}")
        resp.close()
    else:
        # Write response content to file
        target.parent.mkdir(parents=True, exist_ok=True)
        target.write_bytes(resp.content)

    with zipfile.ZipFile(target) as zf:
        # Unpack the entire archive
        zf.extractall(target.parent)
        # The first name list is the top-level directory within the file
        subdir = zf.namelist()[0]

    return target.parent.joinpath(subdir)


def _handle_validate_args(
    schema_dir: Optional[Path], version: Union[str, Version]
) -> Tuple[Path, Version]:
    """Handle arguments for :func:`.install_schemas` and :func:`.validate_xml`."""
    import platformdirs

    supported = {Version["2.1"], Version["3.0.0"]}
    try:
        version = Version[version] if isinstance(version, str) else version
        assert version in supported
    except (AssertionError, KeyError):
        raise NotImplementedError(
            f"SDMX-ML version must be one of {supported}; got {version}"
        ) from None

    # If the user has no preference, download the schemas to the local cache directory
    if not schema_dir:
        schema_dir = platformdirs.user_cache_path("sdmx") / version.name
    schema_dir.mkdir(exist_ok=True, parents=True)

    return schema_dir, version


def install_schemas(
    schema_dir: Optional[Path] = None,
    version: Union[str, Version] = Version["2.1"],
) -> Path:
    """Install SDMX-ML XML Schema documents for use with :func:`.validate_xml`.

    Parameters
    ----------
    schema_dir : Path, optional
        The directory where XSD schemas will be downloaded to. Default: a subdirectory
        named :file:`sdmx/{version}` within the :meth:`platformdirs.user_cache_path`.
    version : str or Version, optional
        The SDMX-ML schema version to install. One of :py:`Version["2.1"]` (default),
        :py:`Version["3.0.0"]`, or :class:`str` equivalent.

    Returns
    -------
    Path
        The path containing the installed schemas. If `schema_dir` is given, the return
        value is identical to the parameter.
    """
    schema_dir, version = _handle_validate_args(schema_dir, version)

    # Copy the entire "schemas" subtree recursively
    copytree(
        _extracted_zipball(version).joinpath("schemas"), schema_dir, dirs_exist_ok=True
    )
    return schema_dir


class XMLFormat:
    NS: Mapping[str, Optional[str]]
    _class_tag: List

    def __init__(self, model, base_ns: str, class_tag: Iterable[Tuple[str, str]]):
        from sdmx import message  # noqa: F401

        self.base_ns = base_ns

        # Construct name spaces
        self.NS = {
            prefix: url if url is None else url.format(base_ns)
            for prefix, url in NS.items()
        }

        # Construct class-tag mapping
        self._class_tag = []

        # Defined in this file
        for name in CT1:
            self._class_tag.append((getattr(model, name), self.qname("str", name)))

        # Defined in this file + those passed to the constructor
        for expr, tag in chain(CT2, class_tag):
            self._class_tag.append((eval(expr), self.qname(tag)))

    @lru_cache()
    def ns_prefix(self, url) -> str:
        """Return the namespace prefix from :attr:`.NS` given its full `url`."""
        for prefix, _url in self.NS.items():
            if url == _url:
                return prefix
        raise ValueError(url)

    @lru_cache()
    def qname(self, ns_or_name, name=None) -> QName:
        """Return a fully-qualified tag `name` in namespace `ns`."""
        if isinstance(ns_or_name, QName):
            # Already a QName; do nothing
            return ns_or_name
        else:
            if name is None:
                match = re.fullmatch(
                    r"(\{(?P<ns_full>.*)\}|(?P<ns_key>.*):)?(?P<name>.*)", ns_or_name
                )
                assert match
                name = match.group("name")
                if ns_key := match.group("ns_key"):
                    ns = self.NS[ns_key]
                elif ns := match.group("ns_full"):
                    pass
                else:
                    ns = None
            else:
                ns = self.NS[ns_or_name]

            return QName(ns, name)

    @lru_cache()
    def class_for_tag(self, tag) -> Optional[type]:
        """Return a message or model class for an XML tag."""
        qname = self.qname(tag)
        results = map(itemgetter(0), filter(lambda ct: ct[1] == qname, self._class_tag))
        try:
            return next(results)
        except StopIteration:
            return None

    @lru_cache()
    def tag_for_class(self, cls):
        """Return an XML tag for a message or model class."""
        results = map(itemgetter(1), filter(lambda ct: ct[0] is cls, self._class_tag))
        try:
            return next(results)
        except StopIteration:
            return None

    # __eq__ and __hash__ to enable lru_cache()
    def __eq__(self, other):
        return self.base_ns == other.base_ns  # pragma: no cover

    def __hash__(self):
        return hash(self.base_ns)



import re
from operator import itemgetter

import pytest

import sdmx
from sdmx import message
from sdmx.model import v21 as model


@pytest.mark.parametrize(
    "cls",
    (message.Message, message.DataMessage, message.StructureMessage, message.Footer),
)
def test_compare(cls):
    A = cls()
    B = cls()

    assert A.compare(B) is True


class TestStructureMessage:
    def test_add_contains_get(self):
        dsd = model.DataStructureDefinition(id="foo")
        msg = message.StructureMessage()

        # add() stores the object
        msg.add(dsd)
        assert 1 == len(msg.structure)

        # __contains__() works
        assert dsd in msg

        # get() retrieves the object
        assert dsd is msg.get("foo")

        # add() with an object not collected in a StructureMessage raises TypeError
        item = model.Item(id="bar")
        with pytest.raises(TypeError):
            msg.add(item)

        # __contains__() also raises TypeError
        with pytest.raises(TypeError):
            item in msg

        # get() with two objects of the same ID raises ValueError
        msg.add(model.DataflowDefinition(id="foo"))
        with pytest.raises(ValueError):
            msg.get("foo")

    def test_dictlike_attribute_access(self):
        dsd = model.DataStructureDefinition(id="foo")
        msg = message.StructureMessage()
        msg.add(dsd)

        # Attribute access works when added to the default, empty DictLike
        assert msg.structure.foo is dsd

    def test_objects(self):
        """:meth:`.objects` can be used to access a collection according to class."""
        msg = message.StructureMessage()
        cl = model.Codelist(id="foo")
        msg.add(cl)

        assert cl is msg.objects(model.Codelist)["foo"]

        with pytest.raises(TypeError):
            # TypeError is raised for classes not collected in this message
            msg.objects(model.DataSet)


EXPECTED = [
    # Structure messages
    (
        "IPI-2010-A21-structure.xml",
        """<sdmx.StructureMessage>
  <Header>
    id: 'categorisation_1450864290565'
    prepared: '2015-12-23T09:51:30.565000+00:00'
    receiver: <Agency unknown>
    sender: <Agency FR1: Institut national de la statistique et des études économiques>
    source: fr: Banque de données macro-économiques
    test: False
  response: <Response [None]>
  Categorisation (1): CAT_IPI-2010_IPI-2010-A21
  CategoryScheme (1): CLASSEMENT_DATAFLOWS
  Codelist (7): CL_FREQ CL_NAF2_A21 CL_NATURE CL_UNIT CL_AREA CL_TIME_C...
  ConceptScheme (1): CONCEPTS_INSEE
  DataflowDefinition (1): IPI-2010-A21
  DataStructureDefinition (1): IPI-2010-A21""",
    ),
    (
        # This message shows the summarization feature: the DFD list is truncated
        "dataflow.xml",
        """<sdmx.StructureMessage>
  <Header>
    id: 'dataflow_ENQ-CONJ-TRES-IND-PERSP_1450865196042'
    prepared: '2015-12-23T10:06:36.042000+00:00'
    receiver: <Agency unknown>
    sender: <Agency FR1: Institut national de la statistique et des études économiques>
    source: fr: Banque de données macro-économiques
    test: False
  response: <Response [None]>
  DataflowDefinition (663): ACT-TRIM-ANC BPM6-CCAPITAL BPM6-CFINANCIER ...
  DataStructureDefinition (663): ACT-TRIM-ANC BPM6-CCAPITAL BPM6-CFINAN...""",
    ),
    # Data messages
    (
        "sg-xs.xml",
        """<sdmx.DataMessage>
  <Header>
    id: 'Generic'
    prepared: '2010-01-04T16:21:49+01:00'
    sender: <Agency ECB>
    source: """
        """
    test: False
  response: <Response [None]>
  DataSet (1)
  dataflow: <DataflowDefinition (missing id)>
  observation_dimension: <Dimension CURRENCY>""",
    ),
    (
        # This message has two DataSets:
        "action-delete.json",
        """<sdmx.DataMessage>
  <Header>
    id: '62b5f19d-f1c9-495d-8446-a3661ed24753'
    prepared: '2012-11-29T08:40:26+00:00'
    sender: <Agency ECB: European Central Bank>
    source: """
        """
    test: False
  response: <Response [None]>
  DataSet (2)
  dataflow: <DataflowDefinition (missing id)>
  observation_dimension: [<Dimension CURRENCY>]""",
    ),
]


@pytest.mark.parametrize(
    "pattern, expected", EXPECTED, ids=list(map(itemgetter(0), EXPECTED))
)
def test_message_repr(specimen, pattern, expected):
    import requests

    with specimen(pattern) as f:
        msg = sdmx.read_sdmx(f)

    # Attach a response object, as if the Message resulted from a requests query
    msg.response = requests.Response()

    if isinstance(expected, re.Pattern):
        assert expected.fullmatch(repr(msg))
    else:
        # __repr__() and __str__() give the same, expected result
        assert expected == repr(msg) == str(msg)



from importlib import import_module
from itertools import chain

from lxml import etree

from sdmx.format import list_media_types
from sdmx.reader.base import BaseReader

from .v21 import XMLParseError

__all__ = ["XMLParseError"]


class Reader(BaseReader):
    """Reader that detects and dispatches to either v21 or v30."""

    media_types = list_media_types(base="xml")
    suffixes = [".xml"]

    @classmethod
    def detect(cls, content):
        return content.startswith(b"<")

    def read_message(self, source, **kwargs):
        # Create an iterative parser
        events = etree.iterparse(source, events=("start", "end"))

        # Peek at the start event for the first tag
        event, element = next(events)

        # Determine the SDMX-ML version
        version = ""
        for url in element.nsmap.values():
            if "v3_0" in url:
                version = "v30"
                break
            elif "v2_1" in url:
                version = "v21"
                break
        assert version, "Cannot determine SDMX-ML version"

        # - Import and instantiate the reader for this version.
        # - Return the peeked (event, element) to the head of the events iterator.
        return (
            import_module(f"sdmx.reader.xml.{version}")
            .Reader()
            .read_message(None, **kwargs, _events=chain([(event, element)], events))
        )



import pytest

from sdmx.session import Session

from . import has_requests_cache


@pytest.mark.skipif(has_requests_cache, reason="test without requests_cache")
def test_session_without_requests_cache():  # pragma: no cover
    # Passing cache= arguments when requests_cache is not installed triggers a warning
    with pytest.warns(RuntimeWarning):
        Session(cache_name="test")


@pytest.mark.network
def test_session_init_cache(tmp_path):
    # Instantiate a REST object with cache
    cache_name = tmp_path / "pandasdmx_cache"
    s = Session(cache_name=str(cache_name), backend="sqlite")

    # Get a resource
    s.get("https://registry.sdmx.org/ws/rest/dataflow")

    # Test for existence of cache file
    assert cache_name.with_suffix(".sqlite").exists()



"""Information Model classes common to SDMX 2.1 and 3.0.

This module implements many of the classes described in the SDMX-IM specification
('spec'), which is available from:

- https://sdmx.org/?page_id=5008
- https://sdmx.org/wp-content/uploads/SDMX_2-1-1_SECTION_2_InformationModel_201108.pdf

Details of the implementation:

- Python dataclasses and type hinting are used to enforce the types of attributes that
  reference instances of other classes.
- Some classes have convenience attributes not mentioned in the spec, to ease navigation
  between related objects. These are marked “:mod:`sdmx` extension not in the IM.”
- Class definitions are grouped by section of the spec, but these sections occasionally
  appear out of order so that classes are defined before they are referenced by others.

"""
import logging
import sys
from abc import ABC, abstractmethod
from collections import ChainMap
from copy import copy
from dataclasses import InitVar, dataclass, field, fields
from datetime import date, datetime, timedelta
from enum import Enum
from functools import lru_cache
from itertools import product
from operator import attrgetter, itemgetter
from typing import (
    Any,
    ClassVar,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from sdmx.dictlike import DictLikeDescriptor
from sdmx.rest import Resource
from sdmx.util import compare, direct_fields, only

from .internationalstring import (
    DEFAULT_LOCALE,
    InternationalString,
    InternationalStringDescriptor,
)

__all__ = [
    "DEFAULT_LOCALE",
    "InternationalString",
    # In the order they appear in this file
    "ConstrainableArtefact",
    "Annotation",
    "AnnotableArtefact",
    "IdentifiableArtefact",
    "NameableArtefact",
    "VersionableArtefact",
    "MaintainableArtefact",
    "ActionType",
    "ConstraintRoleType",
    "FacetValueType",
    "ExtendedFacetValueType",
    "Item",
    "ItemScheme",
    "FacetType",
    "Facet",
    "Representation",
    "Code",
    "Codelist",
    "ISOConceptReference",
    "Concept",
    "ConceptScheme",
    "Component",
    "ComponentList",
    "Category",
    "CategoryScheme",
    "Categorisation",
    "Contact",
    "Organisation",
    "Agency",
    "OrganisationScheme",
    "AgencyScheme",
    "Structure",
    "StructureUsage",
    "DimensionComponent",
    "Dimension",
    "TimeDimension",
    "DimensionDescriptor",
    "GroupDimensionDescriptor",
    "AttributeRelationship",
    "DimensionRelationship",
    "GroupRelationship",
    "DataAttribute",
    "AttributeDescriptor",
    "KeyValue",
    "TimeKeyValue",
    "AttributeValue",
    "Key",
    "GroupKey",
    "SeriesKey",
    "CodingFormat",
    "Level",
    "HierarchicalCode",
    "ConstraintRole",
    "StartPeriod",
    "EndPeriod",
    "CubeRegion",
    "MetadataTargetRegion",
    "DataConsumer",
    "DataProvider",
    "DataConsumerScheme",
    "DataProviderScheme",
    "Datasource",
    "SimpleDatasource",
    "QueryDatasource",
    "RESTDatasource",
    "ProvisionAgreement",
    "CustomType",
    "CustomTypeScheme",
    "NamePersonalisation",
    "NamePersonalisationScheme",
    "Ruleset",
    "RulesetScheme",
    "Transformation",
    "UserDefinedOperator",
    "UserDefinedOperatorScheme",
    "FromVTLSpaceKey",
    "ToVTLSpaceKey",
    "VTLConceptMapping",
    "VTLDataflowMapping",
    "VTLMappingScheme",
    "TransformationScheme",
]

log = logging.getLogger(__name__)

# Utility classes not specified in the SDMX standard


class _MissingID(str):
    def __str__(self):
        return "(missing id)"

    # Supplied to allow this as a default value for dataclass fields
    def __hash__(self):
        return hash(None)  # pragma: no cover

    def __eq__(self, other):
        return isinstance(other, self.__class__)


MissingID = _MissingID()


# §3.2: Base structures


class ConstrainableArtefact:
    """SDMX ConstrainableArtefact."""


@dataclass
class Annotation:
    #: Can be used to disambiguate multiple annotations for one AnnotableArtefact.
    id: Optional[str] = None
    #: Title, used to identify an annotation.
    title: Optional[str] = None
    #: Specifies how the annotation is processed.
    type: Optional[str] = None
    #: A link to external descriptive text.
    url: Optional[str] = None

    #: Content of the annotation.
    text: InternationalStringDescriptor = InternationalStringDescriptor()


@dataclass
class AnnotableArtefact:
    #: :class:`Annotations <.Annotation>` of the object.
    #:
    #: :mod:`.sdmx` implementation detail: The IM does not specify the name of this
    #: feature.
    annotations: List[Annotation] = field(default_factory=list)

    def get_annotation(self, **attrib):
        """Return a :class:`Annotation` with given `attrib`, e.g. 'id'.

        If more than one `attrib` is given, all must match a particular annotation.

        Raises
        ------
        KeyError
            If there is no matching annotation.
        """
        for anno in self.annotations:
            if all(getattr(anno, key, None) == value for key, value in attrib.items()):
                return anno

        raise KeyError(attrib)

    def pop_annotation(self, **attrib):
        """Remove and return a :class:`Annotation` with given `attrib`, e.g. 'id'.

        If more than one `attrib` is given, all must match a particular annotation.

        Raises
        ------
        KeyError
            If there is no matching annotation.
        """
        for i, anno in enumerate(self.annotations):
            if all(getattr(anno, key, None) == value for key, value in attrib.items()):
                return self.annotations.pop(i)

        raise KeyError(attrib)

    def eval_annotation(self, id: str, globals=None):
        """Retrieve the annotation with the given `id` and :func:`eval` its contents.

        This can be used for unpacking Python values (e.g. :class:`dict`) stored as an
        annotation on an AnnotableArtefact (e.g. :class:`~sdmx.model.Code`).

        Returns :obj:`None` if no attribute exists with the given `id`.
        """
        try:
            value = str(self.get_annotation(id=id).text)
        except KeyError:  # No such attribute
            return None

        try:
            return eval(value, globals or {})
        except Exception as e:  # Something that can't be eval()'d, e.g. a plain string
            log.debug(f"Could not eval({value!r}): {e}")
            return value


@dataclass
class IdentifiableArtefact(AnnotableArtefact):
    #: Unique identifier of the object.
    id: str = MissingID
    #: Universal resource identifier that may or may not be resolvable.
    uri: Optional[str] = None
    #: Universal resource name. For use in SDMX registries; all registered objects have
    #: a URN.
    urn: Optional[str] = None

    urn_group: Dict = field(default_factory=dict, repr=False)

    def __post_init__(self):
        if not isinstance(self.id, str):
            raise TypeError(
                f"IdentifiableArtefact.id must be str; got {type(self.id).__name__}"
            )

        if self.urn:
            import sdmx.urn

            self.urn_group = sdmx.urn.match(self.urn)

        try:
            if self.id not in (self.urn_group["item_id"] or self.urn_group["id"]):
                raise ValueError(f"ID {self.id} does not match URN {self.urn}")
        except KeyError:
            pass

    def __eq__(self, other):
        """Equality comparison.

        IdentifiableArtefacts can be compared to other instances. For convenience, a
        string containing the object's ID is also equal to the object.
        """
        if isinstance(other, self.__class__):
            return self.id == other.id
        elif isinstance(other, str):
            return self.id == other

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two IdentifiableArtefacts are the same if they have the same :attr:`id`,
        :attr:`uri`, and :attr:`urn`.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return (
            compare("id", self, other, strict)
            and compare("uri", self, other, strict)
            and compare("urn", self, other, strict)
        )

    def __hash__(self):
        return id(self) if self.id == MissingID else hash(self.id)

    def __lt__(self, other):
        return (
            self.id < other.id if isinstance(other, self.__class__) else NotImplemented
        )

    def __str__(self):
        return self.id

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.id}>"

    @classmethod
    def _preserve(cls, *names: str):
        """Copy dunder `names` from IdentifiableArtefact to a decorated class."""

        def decorator(other_cls):
            for name in map(lambda s: f"__{s}__", names):
                candidates = filter(None, map(lambda k: getattr(k, name), cls.__mro__))
                setattr(other_cls, name, next(candidates))
            return other_cls

        return decorator


@dataclass
@IdentifiableArtefact._preserve("eq", "post_init")
class NameableArtefact(IdentifiableArtefact):
    #: Multi-lingual name of the object.
    name: InternationalStringDescriptor = InternationalStringDescriptor()
    #: Multi-lingual description of the object.
    description: InternationalStringDescriptor = InternationalStringDescriptor()

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two NameableArtefacts are the same if:

        - :meth:`.IdentifiableArtefact.compare` is :obj:`True`, and
        - they have the same :attr:`name` and :attr:`description`.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare` and :meth:`.IdentifiableArtefact.compare`.
        """
        if not super().compare(other, strict):
            pass
        elif self.name != other.name:
            log.debug(
                f"Not identical: name <{repr(self.name)}> != <{repr(other.name)}>"
            )
        elif self.description != other.description:
            log.debug(
                f"Not identical: description <{repr(self.description)}> != "
                f"<{repr(other.description)}>"
            )
        else:
            return True
        return False

    def _repr_kw(self) -> MutableMapping[str, str]:
        name = self.name.localized_default()
        return dict(
            cls=self.__class__.__name__, id=self.id, name=f": {name}" if name else ""
        )

    def __repr__(self) -> str:
        return "<{cls} {id}{name}>".format(**self._repr_kw())


@dataclass
class VersionableArtefact(NameableArtefact):
    #: A version string following an agreed convention.
    version: Optional[str] = None
    #: Date from which the version is valid.
    valid_from: Optional[str] = None
    #: Date from which the version is superseded.
    valid_to: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()
        try:
            if self.version and self.version != self.urn_group["version"]:
                raise ValueError(
                    f"Version {self.version} does not match URN {self.urn}"
                )
            else:
                self.version = self.urn_group["version"]
        except KeyError:
            pass

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two VersionableArtefacts are the same if:

        - :meth:`.NameableArtefact.compare` is :obj:`True`, and
        - they have the same :attr:`version`.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare` and :meth:`.NameableArtefact.compare`.
        """
        return super().compare(other, strict) and compare(
            "version", self, other, strict
        )

    def _repr_kw(self) -> MutableMapping[str, str]:
        return ChainMap(
            super()._repr_kw(),
            dict(version=f"({self.version})" if self.version else ""),
        )


@dataclass
class MaintainableArtefact(VersionableArtefact):
    #: True if the object is final; otherwise it is in a draft state.
    is_final: Optional[bool] = None
    #: :obj:`True` if the content of the object is held externally; i.e., not
    #: the current :class:`Message`.
    is_external_reference: Optional[bool] = None
    #: URL of an SDMX-compliant web service from which the object can be retrieved.
    service_url: Optional[str] = None
    #: URL of an SDMX-ML document containing the object.
    structure_url: Optional[str] = None
    #: Association to the Agency responsible for maintaining the object.
    maintainer: Optional["Agency"] = None

    def __post_init__(self):
        super().__post_init__()
        try:
            if self.maintainer and self.maintainer.id != self.urn_group["agency"]:
                raise ValueError(
                    f"Maintainer {self.maintainer} does not match URN {self.urn}"
                )
            else:
                self.maintainer = Agency(id=self.urn_group["agency"])
        except KeyError:
            pass

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two MaintainableArtefacts are the same if:

        - :meth:`.VersionableArtefact.compare` is :obj:`True`, and
        - they have the same :attr:`maintainer`.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare` and :meth:`.VersionableArtefact.compare`.
        """
        return super().compare(other, strict) and compare(
            "maintainer", self, other, strict
        )

    def _repr_kw(self) -> MutableMapping[str, str]:
        return ChainMap(
            super()._repr_kw(),
            dict(maint=f"{self.maintainer}:" if self.maintainer else ""),
        )

    def __repr__(self) -> str:
        return "<{cls} {maint}{id}{version}{name}>".format(**self._repr_kw())


class BaseConstraint(ABC, MaintainableArtefact):
    """ABC for SDMX 2.1 and 3.0 Constraint."""

    @abstractmethod
    def __contains__(self, name):
        ...


# §3.4: Data Types


ActionType = Enum("ActionType", "delete replace append information")

ConstraintRoleType = Enum("ConstraintRoleType", "allowable actual")

#: SDMX FacetValueType.
#:
#: In the SDMX 2.0 IM, three diagrams in the spec show this enumeration containing
#: 'gregorianYearMonth' but not 'gregorianYear' or 'gregorianMonth'. The table in
#: §3.6.3.3 Representation Constructs does the opposite. One ESTAT query (via SGR) shows
#: a real-world usage of 'gregorianYear'; while one query shows usage of
#: 'gregorianYearMonth'; so all three are included.
FacetValueType = Enum(
    "FacetValueType",
    """string bigInteger integer long short decimal float double boolean uri count
    inclusiveValueRange alpha alphaNumeric numeric exclusiveValueRange incremental
    observationalTimePeriod standardTimePeriod basicTimePeriod gregorianTimePeriod
    gregorianYear gregorianMonth gregorianYearMonth gregorianDay reportingTimePeriod
    reportingYear reportingSemester reportingTrimester reportingQuarter reportingMonth
    reportingWeek reportingDay dateTime timesRange month monthDay day time duration
    keyValues identifiableReference dataSetReference """
    # SDMX 3.0 only
    "geospatialInformation",
)


#: SDMX ExtendedFaceValueType.
#:
#: This enumeration is identical to :class:`.FacetValueType` except for one additional
#: member, "Xhtml". This member is used only in metadata.
ExtendedFacetValueType = Enum(
    "ExtendedFacetValueType",
    """string bigInteger integer long short decimal float double boolean uri count
    inclusiveValueRange alpha alphaNumeric numeric exclusiveValueRange incremental
    observationalTimePeriod standardTimePeriod basicTimePeriod gregorianTimePeriod
    gregorianYear gregorianMonth gregorianYearMonth gregorianDay reportingTimePeriod
    reportingYear reportingSemester reportingTrimester reportingQuarter reportingMonth
    reportingWeek reportingDay dateTime timesRange month monthDay day time duration
    keyValues identifiableReference dataSetReference Xhtml""",
)


UsageStatus = Enum("UsageStatus", "mandatory conditional")


# §3.5: Item Scheme

IT = TypeVar("IT", bound="Item")


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class Item(NameableArtefact, Generic[IT]):
    parent: Optional[Union[IT, "ItemScheme"]] = None
    child: List[IT] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()

        try:
            # Add this Item as a child of its parent
            self.parent.append_child(self)
        except AttributeError:
            pass  # No parent

        # Add this Item as a parent of its children
        for c in self.child:
            c.parent = self

    def __contains__(self, item):
        """Recursive containment."""
        for c in self.child:
            if item == c or item in c:
                return True

    def __iter__(self, recurse=True):
        yield self
        for c in self.child:
            yield from iter(c)

    @property
    def hierarchical_id(self):
        """Construct the ID of an Item in a hierarchical ItemScheme.

        Returns, for example, 'A.B.C' for an Item with id 'C' that is the child of an
        item with id 'B', which is the child of a root Item with id 'A'.

        See also
        --------
        .ItemScheme.get_hierarchical
        """
        return (
            f"{self.parent.hierarchical_id}.{self.id}"
            if isinstance(self.parent, self.__class__)
            else self.id
        )

    def append_child(self, other: IT):
        if other not in self.child:
            self.child.append(other)
        other.parent = self

    def get_child(self, id) -> IT:
        """Return the child with the given *id*."""
        for c in self.child:
            if c.id == id:
                return c
        raise ValueError(id)

    def get_scheme(self):
        """Return the :class:`ItemScheme` to which the Item belongs, if any."""
        try:
            # Recurse
            return self.parent.get_scheme()
        except AttributeError:
            # Either this Item is a top-level Item whose .parent refers to the
            # ItemScheme, or it has no parent
            return self.parent


@dataclass
class ItemScheme(MaintainableArtefact, Generic[IT]):
    """SDMX-IM Item Scheme.

    The IM states that ItemScheme “defines a *set* of :class:`Items <.Item>`…” To
    simplify indexing/retrieval, this implementation uses a :class:`dict` for the
    :attr:`items` attribute, in which the keys are the :attr:`~.IdentifiableArtefact.id`
    of the Item.

    Because this may change in future versions, user code should not access
    :attr:`items` directly. Instead, use the :func:`getattr` and indexing features of
    ItemScheme, or the public methods, to access and manipulate Items:

    >>> foo = ItemScheme(id='foo')
    >>> bar = Item(id='bar')
    >>> foo.append(bar)
    >>> foo
    <ItemScheme: 'foo', 1 items>
    >>> (foo.bar is bar) and (foo['bar'] is bar) and (bar in foo)
    True

    """

    # TODO add delete()
    # TODO add sorting capability; perhaps sort when new items are inserted

    # NB the IM does not specify; this could be True by default, but would need to check
    # against the automatic construction in .reader.*.
    is_partial: Optional[bool] = None

    #: Members of the ItemScheme. Both ItemScheme and Item are abstract classes.
    #: Concrete classes are paired: for example, a :class:`.Codelist` contains
    #: :class:`Codes <.Code>`.
    items: Dict[str, IT] = field(default_factory=dict)

    # The type of the Items in the ItemScheme. This is necessary because the type hint
    # in the class declaration is static; not meant to be available at runtime.
    _Item: ClassVar[Type] = Item

    # Convenience access to items
    def __getattr__(self, name: str) -> IT:
        # Provided to pass test_dsd.py
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError(name)

    def __getitem__(self, name: str) -> IT:
        return self.items[name]

    def get_hierarchical(self, id: str) -> IT:
        """Get an Item by its :attr:`~.Item.hierarchical_id`."""
        if "." not in id:
            return self.items[id]
        else:
            for item in self.items.values():
                if item.hierarchical_id == id:
                    return item
        raise KeyError(id)

    def __contains__(self, item: Union[str, IT]) -> bool:
        """Check containment.

        No recursive search on children is performed as these are assumed to be included
        in :attr:`items`. Allow searching by Item or its id attribute.
        """
        if isinstance(item, str):
            return item in self.items
        return item in self.items.values()

    def __iter__(self):
        return iter(self.items.values())

    def extend(self, items: Iterable[IT]):
        """Extend the ItemScheme with members of `items`.

        Parameters
        ----------
        items : iterable of :class:`.Item`
            Elements must be of the same class as :attr:`items`.
        """
        for i in items:
            self.append(i)

    def __len__(self):
        return len(self.items)

    def append(self, item: IT):
        """Add *item* to the ItemScheme.

        Parameters
        ----------
        item : same class as :attr:`items`
            Item to add.
        """
        if item.id in self.items:
            raise ValueError(f"Item with id {repr(item.id)} already exists")
        self.items[item.id] = item
        if item.parent is None:
            item.parent = self

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two ItemSchemes are the same if:

        - :meth:`.MaintainableArtefact.compare` is :obj:`True`, and
        - their :attr:`items` have the same keys, and corresponding
          :class:`Items <Item>` compare equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare` and :meth:`.MaintainableArtefact.compare`.
        """
        if not super().compare(other, strict):
            pass
        elif set(self.items) != set(other.items):
            log.debug(
                f"ItemScheme contents differ: {repr(set(self.items))} != "
                + repr(set(other.items))
            )
        else:
            for id, item in self.items.items():
                if not item.compare(other.items[id], strict):
                    log.debug(f"…for items with id={repr(id)}")
                    return False
            return True

        return False

    def __repr__(self):
        return "<{cls} {maint}{id}{version} ({N} items){name}>".format(
            **self._repr_kw(), N=len(self.items)
        )

    def setdefault(self, obj=None, **kwargs) -> IT:
        """Retrieve the item *name*, or add it with *kwargs* and return it.

        The returned object is a reference to an object in the ItemScheme, and is of the
        appropriate class.
        """
        if obj and len(kwargs):
            raise ValueError(
                "cannot give both *obj* and keyword arguments to setdefault()"
            )

        if not obj:
            # Replace a string 'parent' ID with a reference to the object
            parent = kwargs.pop("parent", None)
            if isinstance(parent, str):
                kwargs["parent"] = self[parent]

            # Instantiate an object of the correct class
            obj = self._Item(**kwargs)

        try:
            # Add the object to the ItemScheme
            self.append(obj)
        except ValueError:
            # Already present; return the existing object, discard the candidate
            return self[obj.id]
        else:
            return obj


# §3.6: Structure


@dataclass
class FacetType:
    #:
    is_sequence: Optional[bool] = None
    #:
    min_length: Optional[int] = None
    #:
    max_length: Optional[int] = None
    #:
    min_value: Optional[float] = None
    #:
    max_value: Optional[float] = None
    #:
    start_value: Optional[float] = None
    #:
    end_value: Optional[str] = None
    #:
    interval: Optional[float] = None
    #:
    time_interval: Optional[timedelta] = None
    #:
    decimals: Optional[int] = None
    #:
    pattern: Optional[str] = None
    #:
    start_time: Optional[datetime] = None
    #:
    end_time: Optional[datetime] = None
    #: SDMX 3.0 only; not present in SDMX 2.1
    sentinel_values: Optional[str] = None

    def __post_init__(self):
        for name in "max_length", "min_length":
            try:
                setattr(self, name, int(getattr(self, name)))
            except TypeError:
                pass


@dataclass
class Facet:
    #:
    type: FacetType = field(default_factory=FacetType)
    #:
    value: Optional[str] = None
    #:
    value_type: Optional[FacetValueType] = None


@dataclass
class Representation:
    #:
    enumerated: Optional[ItemScheme] = None
    #:
    non_enumerated: List[Facet] = field(default_factory=list)

    def __repr__(self):
        return "<{}: {}, {}>".format(
            self.__class__.__name__, self.enumerated, self.non_enumerated
        )


# §4.3: Codelist


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class Code(Item["Code"]):
    """SDMX Code."""


class Codelist(ItemScheme[IT]):
    """SDMX Codelist."""

    _Item = Code


# §4.4: Concept Scheme


@dataclass
class ISOConceptReference:
    #:
    agency: str
    #:
    id: str
    #:
    scheme_id: str


class Concept(Item["Concept"]):
    #:
    core_representation: Optional[Representation] = None
    #:
    iso_concept: Optional[ISOConceptReference] = None


class ConceptScheme(ItemScheme[Concept]):
    _Item = Concept


# SDMX 2.1 §3.3: Basic Inheritance
# SDMX 3.0 §3.6: The Structure Pattern


@dataclass
@IdentifiableArtefact._preserve("hash", "repr")
class Component(IdentifiableArtefact):
    #:
    concept_identity: Optional[Concept] = None
    #:
    local_representation: Optional[Representation] = None

    def __contains__(self, value):
        for repr in [
            getattr(self.concept_identity, "core_representation", None),
            self.local_representation,
        ]:
            enum = getattr(repr, "enumerated", None)
            if enum is not None:
                return value in enum
        raise TypeError("membership not defined for non-enumerated representations")


CT = TypeVar("CT", bound=Component)


@dataclass
class ComponentList(IdentifiableArtefact, Generic[CT]):
    #:
    components: List[CT] = field(default_factory=list)
    #: Counter used to automatically populate :attr:`.DimensionComponent.order` values.
    auto_order = 1

    # The default type of the Components in the ComponentList. See comment on
    # ItemScheme._Item
    _Component: ClassVar[Type] = Component

    # Convenience access to the components
    def append(self, value: CT) -> None:
        """Append *value* to :attr:`components`."""
        if hasattr(value, "order") and value.order is None:
            value.order = max(self.auto_order, len(self.components) + 1)
            self.auto_order = value.order + 1
        self.components.append(value)

    def extend(self, values: Iterable[CT]) -> None:
        """Extend :attr:`components` with *values*."""
        for value in values:
            self.append(value)

    def get(self, id) -> CT:
        """Return the component with the given *id*."""
        # Search for an existing Component
        for c in self.components:
            if c.id == id:
                return c
        raise KeyError(id)

    def getdefault(self, id, cls=None, **kwargs) -> CT:
        """Return or create the component with the given *id*.

        If the component is automatically created, its :attr:`.Dimension.order`
        attribute is set to the value of :attr:`auto_order`, which is then incremented.

        Parameters
        ----------
        id : str
            Component ID.
        cls : type, optional
            Hint for the class of a new object.
        kwargs
            Passed to the constructor of :class:`.Component`, or a Component subclass if
            :attr:`.components` is overridden in a subclass of ComponentList.
        """
        try:
            return self.get(id)
        except KeyError:
            pass  # No match

        # Create a new object of a class:
        # 1. Given by the cls argument,
        # 2. Specified by a subclass' _default_type attribute, or
        # 3. Hinted for a subclass' components attribute.
        cls = cls or self._Component
        component = cls(id=id, **kwargs)

        if "order" not in kwargs and hasattr(component, "order"):
            # For automatically created dimensions, give a serial value to the order
            # property
            component.order = self.auto_order
            self.auto_order += 1

        self.components.append(component)
        return component

    # Properties of components
    def __getitem__(self, key) -> CT:
        """Convenience access to components."""
        return self.components[key]

    def __len__(self):
        return len(self.components)

    def __iter__(self):
        return iter(self.components)

    def __repr__(self):
        return "<{}: {}>".format(
            self.__class__.__name__, "; ".join(map(repr, self.components))
        )

    def __eq__(self, other):
        """ID equal and same components occur in same order."""
        return super().__eq__(other) and all(
            s == o for s, o in zip(self.components, other.components)
        )

    # Must be reset because __eq__ is defined
    def __hash__(self):
        return super().__hash__()

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two ComponentLists are the same if:

        - :meth:`.IdentifiableArtefact.compare` is :obj:`True`, and
        - corresponding :attr:`components` compare equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare` and :meth:`.IdentifiableArtefact.compare`.
        """
        return super().compare(other, strict) and all(
            c.compare(other.get(c.id), strict) for c in self.components
        )


# §4.5: Category Scheme


class Category(Item["Category"]):
    """SDMX-IM Category."""


class CategoryScheme(ItemScheme[Category]):
    _Item = Category


@dataclass
class Categorisation(MaintainableArtefact):
    #:
    category: Optional[Category] = None
    #:
    artefact: Optional[IdentifiableArtefact] = None


# §4.6: Organisations


@dataclass
class Contact:
    """Organization contact information.

    IMF is the only known data provider that returns messages with :class:`Contact`
    information. These differ from the IM in several ways. This class reflects these
    differences:

    - 'name' and 'org_unit' are InternationalString, instead of strings.
    - 'email' may be a list of e-mail addresses, rather than a single address.
    - 'fax' may be a list of fax numbers, rather than a single number.
    - 'uri' may be a list of URIs, rather than a single URI.
    - 'x400' may be a list of strings, rather than a single string.
    """

    #:
    name: InternationalStringDescriptor = InternationalStringDescriptor()
    #:
    org_unit: InternationalStringDescriptor = InternationalStringDescriptor()
    #:
    telephone: Optional[str] = None
    #:
    responsibility: InternationalStringDescriptor = InternationalStringDescriptor()
    #:
    email: List[str] = field(default_factory=list)
    #:
    fax: List[str] = field(default_factory=list)
    #:
    uri: List[str] = field(default_factory=list)
    #:
    x400: List[str] = field(default_factory=list)


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class Organisation(Item["Organisation"]):
    #:
    contact: List[Contact] = field(default_factory=list)


class OrganisationScheme(ItemScheme[IT]):
    """SDMX OrganisationScheme (abstract class)."""

    _Item = Organisation


# @NameableArtefact._preserve("repr")
class Agency(Organisation):
    """SDMX-IM Organization.

    This class is identical to its parent class.
    """


class AgencyScheme(OrganisationScheme[Agency]):
    _Item = Agency


class DataConsumer(Organisation, ConstrainableArtefact):
    """SDMX DataConsumer."""


class DataConsumerScheme(OrganisationScheme[DataConsumer]):
    _Item = DataConsumer


class DataProvider(Organisation, ConstrainableArtefact):
    """SDMX DataProvider."""


class DataProviderScheme(OrganisationScheme[DataProvider]):
    _Item = DataProvider


# SDMX 3.0 §5.3: Data Structure Definition


@dataclass(repr=False)
class Structure(MaintainableArtefact):
    @property
    def grouping(self) -> Sequence[ComponentList]:
        """A collection of all the ComponentLists associated with a subclass."""
        result: List[ComponentList] = []
        for f in fields(self):
            types = get_args(f.type) or (f.type,)
            try:
                if any(issubclass(t, ComponentList) for t in types):
                    result.append(getattr(self, f.name))
            except TypeError:
                pass
        return result

    def replace_grouping(self, cl: ComponentList) -> None:
        """Replace existing component list with `cl`."""
        field = None
        for f in fields(self):
            is_dictlike = get_origin(f.type) is DictLikeDescriptor
            if f.type == type(cl) or (is_dictlike and get_args(f.type)[1] is type(cl)):
                field = f
                break

        if not field:
            raise TypeError(f"No grouping of type {type(cl)} on {type(self)}")

        if is_dictlike:
            getattr(self, field.name).setdefault(cl.id, cl)
        else:
            setattr(self, field.name, cl)

    def compare(self, other: "Structure", strict: bool = True) -> bool:
        # DictLike of ComponentList will not have an "id" attribute
        def _key(item) -> str:
            return getattr(item, "id", str(type(item)))

        return all(
            s.compare(o, strict)
            for s, o in zip(
                sorted(self.grouping, key=_key), sorted(other.grouping, key=_key)
            )
        )


class StructureUsage(MaintainableArtefact):
    #:
    structure: Optional[Structure] = None


@dataclass
@IdentifiableArtefact._preserve("eq", "hash", "repr")
class DimensionComponent(Component):
    """SDMX DimensionComponent (abstract class)."""

    #:
    order: Optional[int] = None


@dataclass
@IdentifiableArtefact._preserve("eq", "hash", "repr")
class Dimension(DimensionComponent):
    """SDMX Dimension."""

    #:
    concept_role: Optional[Concept] = None


class TimeDimension(DimensionComponent):
    """SDMX TimeDimension."""


@dataclass
class DimensionDescriptor(ComponentList[DimensionComponent]):
    """Describes a set of dimensions.

    IM: “An ordered set of metadata concepts that, combined, classify a statistical
    series, and whose values, when combined (the key) in an instance such as a data set,
    uniquely identify a specific observation.”

    :attr:`.components` is a :class:`list` (ordered) of :class:`Dimension`,
    :class:`MeasureDimension`, and/or :class:`TimeDimension`.
    """

    _Component = Dimension

    def __post_init__(self):
        try:
            # Sort components by already assigned 'order' attributes
            self.components = sorted(self.components, key=lambda c: c.order)
        except TypeError:
            pass  # Some or all of the order attributes are None

    def assign_order(self):
        """Assign the :attr:`.DimensionComponent.order` attribute.

        The Dimensions in :attr:`components` are numbered, starting from 1.
        """
        for i, component in enumerate(self.components):
            component.order = i + 1

    def order_key(self, key):
        """Return a key ordered according to the DSD."""
        result = key.__class__()
        for dim in sorted(self.components, key=attrgetter("order")):
            try:
                result[dim.id] = key[dim.id]
            except KeyError:
                continue
        return result

    @classmethod
    def from_key(cls, key):
        """Create a new DimensionDescriptor from a *key*.

        For each :class:`KeyValue` in the *key*:

        - A new :class:`Dimension` is created.
        - A new :class:`Codelist` is created, containing the
          :attr:`KeyValue.value`.

        Parameters
        ----------
        key : :class:`Key` or :class:`GroupKey` or :class:`SeriesKey`
        """
        dd = cls()
        for order, (id, kv) in enumerate(key.values.items()):
            cl = Codelist(id=id)
            cl.append(Code(id=str(kv.value)))
            dd.components.append(
                Dimension(
                    id=id,
                    local_representation=Representation(enumerated=cl),
                    order=order,
                )
            )
        return dd


class GroupDimensionDescriptor(DimensionDescriptor):
    #:
    attachment_constraint: Optional[bool] = None
    # #:
    # constraint: Optional[AttachmentConstraint] = None

    def assign_order(self):
        """:meth:`assign_order` has no effect for GroupDimensionDescriptor."""
        pass


class AttributeRelationship:
    pass


@dataclass
class DimensionRelationship(AttributeRelationship):
    #:
    dimensions: List[DimensionComponent] = field(default_factory=list)
    #: NB the IM says "0..*" here in a diagram, but the text does not match.
    group_key: Optional["GroupDimensionDescriptor"] = None


@dataclass
class GroupRelationship(AttributeRelationship):
    #: “Retained for compatibility reasons” in SDMX 2.1 versus 2.0; not used by
    #: :mod:`sdmx`.
    group_key: Optional["GroupDimensionDescriptor"] = None


@dataclass
@NameableArtefact._preserve("eq", "hash")
class DataAttribute(Component):
    #:
    related_to: Optional[AttributeRelationship] = None
    #:
    usage_status: Optional[UsageStatus] = None
    #:
    concept_role: Optional[Concept] = None


class AttributeDescriptor(ComponentList[DataAttribute]):
    _Component = DataAttribute


@dataclass(repr=False)
@IdentifiableArtefact._preserve("hash")
class BaseDataStructureDefinition(Structure, ConstrainableArtefact):
    """Common features of SDMX 2.1 and 3.0 DataStructureDefinition (**DSD**)."""

    #: A :class:`AttributeDescriptor` that describes the attributes of the data
    #: structure.
    attributes: AttributeDescriptor = field(default_factory=AttributeDescriptor)
    #: A :class:`DimensionDescriptor` that describes the dimensions of the data
    #: structure.
    dimensions: DimensionDescriptor = field(default_factory=DimensionDescriptor)
    #: Mapping from  :attr:`.GroupDimensionDescriptor.id` to
    #: :class:`.GroupDimensionDescriptor`.
    group_dimensions: DictLikeDescriptor[
        str, GroupDimensionDescriptor
    ] = DictLikeDescriptor()

    # Specific types to be used in concrete subclasses
    MemberValue: ClassVar[Type["BaseMemberValue"]]
    MemberSelection: ClassVar[Type["BaseMemberSelection"]]
    ConstraintType: ClassVar[Type[BaseConstraint]]

    # Convenience methods
    def iter_keys(
        self, constraint: Optional[BaseConstraint] = None, dims: List[str] = []
    ) -> Generator["Key", None, None]:
        """Iterate over keys.

        Parameters
        ----------
        constraint : :class:`Constraint <.BaseConstraint>`, optional
            If given, only yield Keys that are within the constraint.
        dims : list of str, optional
            If given, only iterate over allowable values for the Dimensions with these
            IDs. Other dimensions have only a single value like "(DIM_ID)", where
            DIM_ID is the ID of the dimension.
        """
        # NB for performance, the implementation tries to use iterators and avoid
        #    constructing full-length tuples/lists at any point

        _constraint = constraint or NullConstraint
        dims = dims or [dim.id for dim in self.dimensions.components]

        # Utility to return an immutable function that produces KeyValues. The
        # arguments are frozen so these can be set using loop variables and stored in a
        # map() object that isn't modified on future loops
        def make_factory(id=None, value_for=None):
            return lambda value: KeyValue(id=id, value=value, value_for=value_for)

        # List of iterables of (dim.id, KeyValues) along each dimension
        all_kvs: List[Iterable[KeyValue]] = []

        # Iterate over dimensions
        for dim in self.dimensions.components:
            if (
                dim.id not in dims
                or dim.local_representation is None
                or dim.local_representation.enumerated is None
            ):
                # `dim` is not enumerated by an ItemScheme, or not included in the
                # `dims` argument and not to be iterated over. Create a placeholder.
                all_kvs.append(
                    [KeyValue(id=dim.id, value=f"({dim.id})", value_for=dim)]
                )
            else:
                # Create a KeyValue for each Item in the ItemScheme; filter through any
                # constraint.
                all_kvs.append(
                    filter(
                        _constraint.__contains__,
                        map(
                            make_factory(id=dim.id, value_for=dim),
                            dim.local_representation.enumerated,
                        ),
                    ),
                )

        # Create Key objects from Cartesian product of KeyValues along each dimension
        # NB this does not work with DataKeySet
        # TODO improve to work with DataKeySet
        yield from filter(_constraint.__contains__, map(Key, product(*all_kvs)))

    def make_constraint(self, key):
        """Return a constraint for `key`.

        `key` is a :class:`dict` wherein:

        - keys are :class:`str` ids of Dimensions appearing in this DSD's
          :attr:`dimensions`, and
        - values are '+'-delimited :class:`str` containing allowable values, *or*
          iterables of :class:`str`, each an allowable value.

        For example::

            cc2 = dsd.make_constraint({'foo': 'bar+baz', 'qux': 'q1+q2+q3'})

        ``cc2`` includes any key where the 'foo' dimension is 'bar' *or* 'baz', *and*
        the 'qux' dimension is one of 'q1', 'q2', or 'q3'.

        Returns
        -------
        ContentConstraint
            A constraint with one :class:`CubeRegion` in its
            :attr:`data_content_region <ContentConstraint.data_content_region>` ,
            including only the values appearing in `key`.

        Raises
        ------
        ValueError
            if `key` contains a dimension IDs not appearing in :attr:`dimensions`.
        """
        # Make a copy to avoid pop()'ing off the object in the calling scope
        key = key.copy()

        cr = CubeRegion()
        for dim in self.dimensions:
            mvs = set()
            try:
                values = key.pop(dim.id)
            except KeyError:
                continue

            values = values.split("+") if isinstance(values, str) else values
            for value in values:
                # TODO validate values
                mvs.add(self.MemberValue(value=value))

            cr.member[dim] = self.MemberSelection(
                included=True, values_for=dim, values=mvs
            )

        if len(key):
            raise ValueError(
                "Dimensions {!r} not in {!r}".format(list(key.keys()), self.dimensions)
            )

        return self.ConstraintType(
            data_content_region=[cr],
            role=ConstraintRole(role=ConstraintRoleType.allowable),
        )

    @classmethod
    def from_keys(cls, keys):
        """Return a new DSD given some *keys*.

        The DSD's :attr:`dimensions` refers to a set of new :class:`Concepts <Concept>`
        and :class:`Codelists <Codelist>`, created to represent all the values observed
        across *keys* for each dimension.

        Parameters
        ----------
        keys : iterable of :class:`Key`
            or of subclasses such as :class:`SeriesKey` or :class:`GroupKey`.
        """
        iter_keys = iter(keys)
        dd = DimensionDescriptor.from_key(next(iter_keys))

        for k in iter_keys:
            for i, (id, kv) in enumerate(k.values.items()):
                try:
                    dd[i].local_representation.enumerated.append(Code(id=str(kv.value)))
                except ValueError:
                    pass  # Item already exists

        return cls(dimensions=dd)

    def make_key(self, key_cls, values: Mapping, extend=False, group_id=None):
        """Make a :class:`.Key` or subclass.

        Parameters
        ----------
        key_cls : Key or SeriesKey or GroupKey
            Class of Key to create.
        values : dict
            Used to construct :attr:`.Key.values`.
        extend : bool, optional
            If :obj:`True`, make_key will not return :class:`KeyError` on missing
            dimensions. Instead :attr:`dimensions` (`key_cls` is Key or SeriesKey) or
            :attr:`group_dimensions` (`key_cls` is GroupKey) will be extended by
            creating new Dimension objects.
        group_id : str, optional
            When `key_cls` is :class`.GroupKey`, the ID of the
            :class:`.GroupDimensionDescriptor` that structures the key.

        Returns
        -------
        Key
            An instance of `key_cls`.

        Raises
        ------
        KeyError
            If any of the keys of `values` is not a Dimension or Attribute in the DSD.
        """
        # Methods to get dimensions and attributes
        get_method = "getdefault" if extend else "get"
        dim = getattr(self.dimensions, get_method)
        attr = getattr(self.attributes, get_method)

        # Arguments for creating the Key
        args: Dict[str, Any] = dict(described_by=self.dimensions)

        if key_cls is GroupKey:
            # Get the GroupDimensionDescriptor, if indicated by group_id
            gdd = self.group_dimensions.get(group_id, None)

            if group_id and not gdd and not extend:
                # Cannot create
                raise KeyError(group_id)
            elif group_id and extend:
                # Create the GDD
                gdd = GroupDimensionDescriptor(id=group_id)
                self.group_dimensions[gdd.id] = gdd

                # GroupKey will have same ID and be described by the GDD
                args = dict(id=group_id, described_by=gdd)

                # Dimensions to be retrieved from the GDD
                def dim(id):  # noqa: F811
                    # Get from the DimensionDescriptor
                    new_dim = self.dimensions.getdefault(id)
                    # Add to the GDD
                    gdd.components.append(new_dim)
                    return gdd.get(id)

            else:
                # Not described by anything
                args = dict()

        key = key_cls(**args)

        # Convert keyword arguments to either KeyValue or AttributeValue
        keyvalues = []
        for order, (id, value) in enumerate(values.items()):
            if id in self.attributes.components:
                # Reference a DataAttribute from the AttributeDescriptor
                da = attr(id)
                # Store the attribute value, referencing da
                key.attrib[da.id] = AttributeValue(value=value, value_for=da)
                continue

            # Reference a Dimension from the DimensionDescriptor. If extend=False and
            # the Dimension does not exist, this will raise KeyError
            args = dict(id=id, value=value, value_for=dim(id))

            # Retrieve the order
            order = args["value_for"].order

            # Store a KeyValue, to be sorted later
            keyvalues.append((order, KeyValue(**args)))

        # Sort the values according to *order*
        key.values.update({kv.id: kv for _, kv in sorted(keyvalues)})

        return key


@dataclass(repr=False)
class BaseDataflow(StructureUsage, ConstrainableArtefact):
    """Common features of SDMX 2.1 DataflowDefinition and 3.0 Dataflow."""

    structure: BaseDataStructureDefinition = field(
        default_factory=BaseDataStructureDefinition
    )

    def __post_init__(self):
        super().__post_init__()

        # Factory default `structure` inherits is_external_reference from the data flow
        if self.structure.is_external_reference is None:
            self.structure.is_external_reference = self.is_external_reference

    def iter_keys(
        self, constraint: Optional[BaseConstraint] = None, dims: List[str] = []
    ) -> Generator["Key", None, None]:
        """Iterate over keys.

        See also
        --------
        .BaseDataStructureDefinition.iter_keys
        """
        yield from self.structure.iter_keys(constraint=constraint, dims=dims)


class _AllDimensions:
    pass


#: A singleton.
AllDimensions = _AllDimensions()


# §5.4: Data Set


def value_for_dsd_ref(kind, args, kwargs):
    """Maybe replace a string 'value_for' in *kwargs* with a DSD reference."""
    try:
        dsd = kwargs.pop("dsd")
        descriptor = getattr(dsd, kind + "s")
        kwargs["value_for"] = descriptor.get(kwargs["value_for"])
    except KeyError:
        pass
    return args, kwargs


@dataclass
class KeyValue:
    """One value in a multi-dimensional :class:`Key`."""

    #:
    id: str
    #: The actual value.
    value: Any
    #:
    value_for: Optional[DimensionComponent] = None

    dsd: InitVar[BaseDataStructureDefinition] = None

    def __post_init__(self, dsd):
        if dsd:
            self.value_for = getattr(dsd, "dimensions").get(self.value_for)

    def __eq__(self, other):
        """Compare the value to a simple Python built-in type or other key-like.

        `other` may be :class:`.KeyValue` or :class:`.ComponentValue`; if so, and both
        `self` and `other` have :attr:`.value_for`, these must refer to the same object.
        """
        other_value = self._compare_value(other)
        result = self.value == other_value
        if isinstance(other, (KeyValue, ComponentValue)):
            result &= (
                self.value_for in (None, other.value_for) or other.value_for is None
            )
        return result

    @staticmethod
    def _compare_value(other):
        if isinstance(other, (KeyValue, ComponentValue, BaseMemberValue)):
            return other.value
        else:
            return other

    def __lt__(self, other):
        return self.value < self._compare_value(other)

    def __str__(self):
        return f"{self.id}={self.value}"

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.id}={self.value}>"

    def __hash__(self):
        # KeyValue instances with the same id & value hash identically
        return hash(self.id + str(self.value))


class TimeKeyValue(KeyValue):
    """SDMX TimeKeyValue.

    Identical to its parent class.
    """


@dataclass
class AttributeValue:
    """SDMX AttributeValue.

    In the spec, AttributeValue is an abstract class. Here, it serves as both the
    concrete subclasses CodedAttributeValue and UncodedAttributeValue.
    """

    # TODO separate and enforce properties of Coded- and UncodedAttributeValue
    #:
    value: Union[str, Code]
    #:
    value_for: Optional[DataAttribute] = None
    #:
    start_date: Optional[date] = None

    dsd: InitVar[BaseDataStructureDefinition] = None

    def __post_init__(self, dsd):
        if dsd:
            self.value_for = getattr(dsd, "attributes").get(self.value_for)

    def __eq__(self, other):
        """Compare the value to a Python built-in type, e.g. str."""
        return self.value == other

    def __str__(self):
        # self.value directly for UncodedAttributeValue
        return getattr(
            self.value,
            "id",
            self.value if isinstance(self.value, str) else repr(self.value),
        )

    def __repr__(self):
        return "<{}: {}={}>".format(self.__class__.__name__, self.value_for, self.value)

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two AttributeValues are equal if their properties are equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return all(
            compare(attr, self, other, strict)
            for attr in ["start_date", "value", "value_for"]
        )


@dataclass
class Key:
    """SDMX Key class.

    The constructor takes an optional list of keyword arguments; the keywords are used
    as Dimension or Attribute IDs, and the values as KeyValues.

    For convenience, the values of the key may be accessed directly:

    >>> k = Key(foo=1, bar=2)
    >>> k.values['foo']
    1
    >>> k['foo']
    1

    Parameters
    ----------
    dsd : DataStructureDefinition <.BaseDataStructureDefinition>
        If supplied, the :attr:`~.BaseDataStructureDefinition.dimensions` and
        :attr:`~.BaseDataStructureDefinition.attributes` are used to separate the
        `kwargs` into :class:`KeyValues <.KeyValue>` and
        :class:`AttributeValues <.AttributeValue>`. The `kwargs` for
        :attr:`described_by`, if any, must be
        :attr:`~.BaseDataStructureDefinition.dimensions` or appear in
        :attr:`~.BaseDataStructureDefinition.group_dimensions`.
    kwargs
        Dimension and Attribute IDs, and/or the class properties.

    """

    #:
    attrib: DictLikeDescriptor[str, AttributeValue] = DictLikeDescriptor()
    #:
    described_by: Optional[DimensionDescriptor] = None
    #: Individual KeyValues that describe the key.
    values: DictLikeDescriptor[str, KeyValue] = DictLikeDescriptor()

    def __init__(self, arg: Union[Mapping, Sequence[KeyValue], None] = None, **kwargs):
        # Handle kwargs corresponding to attributes
        self.attrib.update(kwargs.pop("attrib", {}))

        # DimensionDescriptor
        dd = kwargs.pop("described_by", None)
        self.described_by = dd

        if arg and isinstance(arg, Mapping):
            if len(kwargs):
                raise ValueError(
                    "Key() accepts either a single argument, or keyword arguments; not "
                    "both."
                )
            kwargs.update(arg)

        kvs: Iterable[Tuple] = []

        if isinstance(arg, Sequence):
            # Sequence of already-prepared KeyValues; assume already sorted
            kvs = map(lambda kv: (kv.id, kv), arg)
        else:
            # Convert bare keyword arguments to KeyValue
            _kvs = []
            for order, (id, value) in enumerate(kwargs.items()):
                args = dict(id=id, value=value)
                if dd:
                    # Reference the Dimension
                    args["value_for"] = dd.get(id)
                    # Use the existing Dimension's order attribute
                    order = args["value_for"].order

                # Store a KeyValue, to be sorted later
                _kvs.append((order, (id, KeyValue(**args))))

            # Sort the values according to *order*, then unwrap
            kvs = map(itemgetter(1), sorted(_kvs))

        self.values.update(kvs)

    def __len__(self):
        """The length of the Key is the number of KeyValues it contains."""
        return len(self.values)

    def __contains__(self, other):
        """A Key contains another if it is a superset."""
        try:
            return all([self.values[k] == v for k, v in other.values.items()])
        except KeyError:
            # 'k' in other does not appear in this Key()
            return False

    def __iter__(self):
        yield from self.values.values()

    # Convenience access to values by name
    def __getitem__(self, name):
        return self.values[name]

    def __setitem__(self, name, value):
        # Convert a bare string or other Python object to a KeyValue instance
        if not isinstance(value, KeyValue):
            value = KeyValue(id=name, value=value)
        self.values[name] = value

    # Convenience access to values by attribute
    def __getattr__(self, name):
        try:
            return self.values[name]
        except KeyError:
            raise AttributeError(name)

    # Copying
    def __copy__(self):
        result = Key()
        if self.described_by:
            result.described_by = self.described_by
        for kv in self.values.values():
            result[kv.id] = kv
        return result

    def copy(self, arg=None, **kwargs):
        result = copy(self)
        for id, value in kwargs.items():
            result[id] = value
        return result

    def __add__(self, other):
        if other is None:
            other_values = dict()
        elif not isinstance(other, Key):
            raise NotImplementedError
        else:
            other_values = other.values
        result = copy(self)
        for id, value in other_values.items():
            result[id] = value
        return result

    def __radd__(self, other):
        if other is None:
            return copy(self)
        else:
            raise NotImplementedError

    def __eq__(self, other):
        if hasattr(other, "values"):
            # Key
            return all(
                [a == b for a, b in zip(self.values.values(), other.values.values())]
            )
        elif hasattr(other, "key_value"):
            # DataKey
            return all(
                [a == b for a, b in zip(self.values.values(), other.key_value.values())]
            )
        elif isinstance(other, str) and len(self.values) == 1:
            return self.values[0] == other
        else:
            raise ValueError(other)

    def __hash__(self):
        # Hash of the individual KeyValues, in order
        return hash(tuple(hash(kv) for kv in self.values.values()))

    # Representations

    def __str__(self):
        return "({})".format(", ".join(map(str, self.values.values())))

    def __repr__(self):
        return "<{}: {}>".format(
            self.__class__.__name__, ", ".join(map(str, self.values.values()))
        )

    def order(self, value=None):
        if value is None:
            value = self
        try:
            return self.described_by.order_key(value)
        except AttributeError:
            return value

    def get_values(self):
        return tuple([kv.value for kv in self.values.values()])


class GroupKey(Key):
    #:
    id: Optional[str] = None
    #:
    described_by: Optional[GroupDimensionDescriptor] = None

    def __init__(self, arg: Optional[Mapping] = None, **kwargs):
        # Remove the 'id' keyword argument
        id = kwargs.pop("id", None)
        super().__init__(arg, **kwargs)
        self.id = id


@dataclass
class SeriesKey(Key):
    #: :mod:`sdmx` extension not in the IM.
    group_keys: Set[GroupKey] = field(default_factory=set)

    __eq__ = Key.__eq__
    __hash__ = Key.__hash__
    __repr__ = Key.__repr__

    @property
    def group_attrib(self):
        """Return a view of attributes on all :class:`GroupKey` including the series."""
        # Needed to pass existing tests
        view = dict()
        for gk in self.group_keys:
            view.update(gk.attrib)
        return view


@dataclass
class BaseObservation:
    """Common features of SDMX 2.1 and 3.0 Observation.

    This class also implements the IM classes ObservationValue, UncodedObservationValue,
    and CodedObservation.
    """

    #:
    attached_attribute: DictLikeDescriptor[str, AttributeValue] = DictLikeDescriptor()
    #:
    series_key: Optional[SeriesKey] = None
    #: Key for dimension(s) varying at the observation level.
    dimension: Optional[Key] = None
    #: Data value.
    value: Optional[Union[Any, Code]] = None
    #: :mod:`sdmx` extension not in the IM.
    group_keys: Set[GroupKey] = field(default_factory=set)

    @property
    def attrib(self):
        """Return a view of combined observation, series & group attributes."""
        view = self.attached_attribute.copy()
        view.update(getattr(self.series_key, "attrib", {}))
        for gk in self.group_keys:
            view.update(gk.attrib)
        return view

    @property
    def dim(self):
        return self.dimension

    @property
    def key(self):
        """Return the entire key, including KeyValues at the series level."""
        return (self.series_key or SeriesKey()) + self.dimension

    def __len__(self):
        # FIXME this is unintuitive; maybe deprecate/remove?
        return len(self.key)

    def __str__(self):
        return "{0.key}: {0.value}".format(self)

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two Observations are equal if:

        - their :attr:`dimension`, :attr:`value`, :attr:`series_key`, and
          :attr:`value_for` are all equal,
        - their corresponding :attr:`attached_attribute` and :attr:`group_keys` are all
          equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return (
            all(
                compare(attr, self, other, strict)
                for attr in ["dimension", "series_key", "value", "value_for"]
            )
            and self.attached_attribute.compare(other.attached_attribute)
            and self.group_keys == other.group_keys
        )


@dataclass
class BaseDataSet(AnnotableArtefact):
    """Common features of SDMX 2.1 and 3.0 DataSet."""

    #:
    action: Optional[ActionType] = None
    #:
    valid_from: Optional[str] = None
    #:
    described_by: Optional[BaseDataflow] = None
    #:
    structured_by: Optional[BaseDataStructureDefinition] = None

    #: All observations in the DataSet.
    obs: List[BaseObservation] = field(default_factory=list)

    #: Map of series key → list of observations.
    #: :mod:`sdmx` extension not in the IM.
    series: DictLikeDescriptor[SeriesKey, List[BaseObservation]] = DictLikeDescriptor()
    #: Map of group key → list of observations.
    #: :mod:`sdmx` extension not in the IM.
    group: DictLikeDescriptor[GroupKey, List[BaseObservation]] = DictLikeDescriptor()

    def __post_init__(self):
        if self.action and not isinstance(self.action, ActionType):
            self.action = ActionType[self.action]

    def __len__(self):
        return len(self.obs)

    def _add_group_refs(self, target):
        """Associate *target* with groups in this dataset.

        *target* may be an instance of SeriesKey or Observation.
        """
        for group_key in self.group:
            if group_key in (target if isinstance(target, SeriesKey) else target.key):
                target.group_keys.add(group_key)
                if isinstance(target, BaseObservation):
                    self.group[group_key].append(target)

    def add_obs(self, observations, series_key=None):
        """Add *observations* to a series with *series_key*.

        Checks consistency and adds group associations."""
        if series_key is not None:
            # Associate series_key with any GroupKeys that apply to it
            self._add_group_refs(series_key)
            # Maybe initialize empty series
            self.series.setdefault(series_key, [])

        for obs in observations:
            # Associate the observation with any GroupKeys that contain it
            self._add_group_refs(obs)

            # Store a reference to the observation
            self.obs.append(obs)

            if series_key is not None:
                if obs.series_key is None:
                    # Assign the observation to the SeriesKey
                    obs.series_key = series_key
                else:
                    # Check that the Observation is not associated with a different
                    # SeriesKey
                    assert obs.series_key is series_key

                # Store a reference to the observation
                self.series[series_key].append(obs)

    def __str__(self):
        return (
            f"<DataSet structured_by={self.structured_by!r} with {len(self)} "
            "observations>"
        )

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two DataSets are the same if:

        - their :attr:`action`, :attr:`valid_from` compare equal.
        - all dataset-level attached attributes compare equal.
        - they have the same number of observations, series, and groups.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return (
            compare("action", self, other, strict)
            and compare("valid_from", self, other, strict)
            and self.attrib.compare(other.attrib, strict)
            and len(self.obs) == len(other.obs)
            and len(self.series) == len(other.series)
            and len(self.group) == len(other.group)
            and all(o[0].compare(o[1], strict) for o in zip(self.obs, other.obs))
        )


# §7.3: Metadata Structure Definition


class AttributeComponent(Component):
    """SDMX 3.0 AttributeComponent.

    .. note:: This intermediate, abstract class is not present in the SDMX 2.1 IM.
    """


@dataclass
class MetadataAttribute(AttributeComponent):
    """SDMX MetadataAttribute."""

    is_presentational: Optional[bool] = None
    max_occurs: Optional[int] = None
    min_occurs: Optional[int] = None

    parent: Optional["MetadataAttribute"] = None
    child: List["MetadataAttribute"] = field(default_factory=list)


class BaseMetadataStructureDefinition(Structure, ConstrainableArtefact):
    """ABC for SDMX 2.1 and 3.0 MetadataStructureDefinition."""


class BaseMetadataflow(StructureUsage, ConstrainableArtefact):
    """ABC for SDMX 2.1 MetadataflowDefinition and SDMX 3.0 Metadataflow."""


# §7.4 MetadataSet


@dataclass
class BaseTextAttributeValue:
    """ABC for SDMX 2.1 and 3.0 TextAttributeValue."""

    text: InternationalStringDescriptor = InternationalStringDescriptor()


@dataclass
class BaseXHTMLAttributeValue:
    """ABC for SDMX 2.1 and 3.0 XHTMLAttributeValue."""

    value: str


@dataclass
class BaseMetadataSet:
    """ABC for SDMX 2.1 and 3.0 MetadataSet."""

    action: Optional[ActionType] = None

    reporting_begin: Optional[date] = None
    reporting_end: Optional[date] = None

    publication_period: Optional[date] = None
    publication_year: Optional[date] = None


# SDMX 2.1 §8: Hierarchical Code List
# SDMX 3.0 §8: Hierarchy


class CodingFormat:
    """SDMX CodingFormat."""

    coding_format: Facet


@dataclass
class Level(NameableArtefact):
    """SDMX Level."""

    parent: Optional[Union["Level", Any]] = None  # NB second element is "Hierarchy"
    child: Optional["Level"] = None

    code_format: CodingFormat = field(default_factory=CodingFormat)


@dataclass
class HierarchicalCode(IdentifiableArtefact):
    """SDMX HierarchicalCode."""

    #: Date from which the construct is valid.
    valid_from: Optional[str] = None
    #: Date from which the construct is superseded.
    valid_to: Optional[str] = None

    #: The Code that is used at the specific point in the hierarchy.
    code: Optional[Code] = None

    level: Optional[Level] = None

    parent: Optional[
        Union["HierarchicalCode", Any]
    ] = None  # NB second element is "Hierarchy"
    child: List["HierarchicalCode"] = field(default_factory=list)


# SDMX 2.1 §10.2: Constraint inheritance
# SDMX 3.0 §12: Constraints


@dataclass
class ConstraintRole:
    #:
    role: ConstraintRoleType


@dataclass
class ComponentValue:
    #:
    value_for: Component
    #:
    value: Any


@dataclass
class TimeDimensionValue(ComponentValue):
    time_value: Any
    operator: str


class BaseSelectionValue:
    """ABC for SDMX 2.1 and 3.0 SelectionValue."""


@dataclass
class BaseMemberValue:
    """Common features of SDMX 2.1 and 3.0 MemberValue."""

    #:
    value: str
    #:
    cascade_values: Optional[bool] = None

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value == (other.value if isinstance(other, KeyValue) else other)

    def __repr__(self):
        return f"{repr(self.value)}" + (" + children" if self.cascade_values else "")


@dataclass
class Period:
    """Class not specified in the IM."""

    is_inclusive: bool
    period: datetime


class StartPeriod(Period):
    pass


class EndPeriod(Period):
    pass


@dataclass
class BaseDataKey:
    """Common features of SDMX 2.1 and 3.0 DataKey."""

    #: :obj:`True` if the :attr:`keys` are included in the :class:`.Constraint`;
    # :obj:`False` if they are excluded.
    included: bool
    #: Mapping from :class:`.Component` to :class:`.ComponentValue` comprising the key.
    key_value: Dict[Component, ComponentValue] = field(default_factory=dict)


@dataclass
class BaseDataKeySet:
    """Common features of SDMX 2.1 and 3.0 DataKeySet."""

    #: :obj:`True` if the :attr:`keys` are included in the
    #: :class:`Constraint <.BaseConstraint>`; :obj:`False` if they are excluded.
    included: bool
    #: :class:`DataKeys <.BaseDataKey>` appearing in the set.
    keys: List[BaseDataKey] = field(default_factory=list)

    def __len__(self):
        """:func:`len` of the DataKeySet = :func:`len` of its :attr:`keys`."""
        return len(self.keys)

    def __contains__(self, item):
        return any(item == dk for dk in self.keys)


@dataclass
class BaseMemberSelection:
    """Common features of SDMX 2.1 and 3.0 MemberSelection."""

    #:
    values_for: Component
    #:
    included: bool = True
    #: Value(s) included in the selection. Note that the name of this attribute is not
    #: stated in the IM, so 'values' is chosen for the implementation in this package.
    values: List[BaseSelectionValue] = field(default_factory=list)

    def __contains__(self, value):
        """Compare KeyValue to MemberValue."""
        return any(mv == value for mv in self.values) is self.included

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.values_for.id} "
            f"{'not ' if not self.included else ''}in {{"
            f"{', '.join(map(repr, self.values))}}}>"
        )


@dataclass
class CubeRegion:
    #:
    included: bool = True
    #:
    member: Dict[DimensionComponent, BaseMemberSelection] = field(default_factory=dict)

    def __contains__(self, other: Union["Key", "KeyValue"]) -> bool:
        """Membership test.

        `other` may be either:

        - :class:`.Key` —all its :class:`.KeyValue` are checked.
        - :class:`.KeyValue` —only the one :class:`.Dimension` for which `other` is a
          value is checked

        Returns
        -------
        bool
            :obj:`True` if:

            - :attr:`.included` *and* `other` is in the CubeRegion;
            - if :attr:`.included` is :obj:`False` *and* `other` is outside the
              CubeRegion; or
            - the `other` is KeyValue referencing a Dimension that is not included in
              :attr:`.member`.
        """
        if isinstance(other, Key):
            result = all(other[ms.values_for.id] in ms for ms in self.member.values())
        elif other.value_for is None:
            # No Dimension reference to use
            result = False
        elif other.value_for not in self.member or len(self.member) > 1:
            # This CubeRegion doesn't have a MemberSelection for the KeyValue's
            # Component; or it concerns additional Components, so inclusion can't be
            # determined
            return True
        else:
            # Check whether the KeyValue is in the indicated dimension
            result = other.value in self.member[other.value_for]

        # Return the correct sense
        return result is self.included

    def to_query_string(self, structure):
        all_values = []

        for dim in structure.dimensions:
            if isinstance(dim, TimeDimension):
                # TimeDimensions handled by query parameters
                continue
            ms = self.member.get(dim, None)
            values = sorted(mv.value for mv in ms.values) if ms else []
            all_values.append("+".join(values))

        return ".".join(all_values)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {'in' if self.included else 'ex'}clude "
            f"{' '.join(map(repr, self.member.values()))}>"
        )


@dataclass
class MetadataTargetRegion:
    #:
    is_included: bool = True


class _NullConstraint:
    """Constraint that allows anything."""

    def __contains__(self, value):
        return True


NullConstraint = _NullConstraint()

# SDMX 2.1 §11: Data Provisioning
# SDMX 3.0 §13: Data Provisioning


class Datasource:
    url: str


class SimpleDatasource(Datasource):
    pass


class QueryDatasource(Datasource):
    # Abstract.
    # NB the SDMX-IM inconsistently uses this name and 'WebServicesDatasource'.
    pass


class RESTDatasource(QueryDatasource):
    pass


@dataclass
@MaintainableArtefact._preserve("hash")
class ProvisionAgreement(MaintainableArtefact, ConstrainableArtefact):
    #:
    structure_usage: Optional[StructureUsage] = None
    #:
    data_provider: Optional[DataProvider] = None


# SDMX 3.0 §15: Validation and Transformation Language


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class CustomType(Item["CustomType"]):
    data_type: Optional[str] = None
    null_value: Optional[str] = None
    output_format: Optional[str] = None
    vtl_literal_format: Optional[str] = None
    vtl_scalar_type: Optional[str] = None


class CustomTypeScheme(ItemScheme[CustomType]):
    _Item = CustomType


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class NamePersonalisation(Item["NamePersonalisation"]):
    vtl_default_name: Optional[str] = None


class NamePersonalisationScheme(ItemScheme[NamePersonalisation]):
    _Item = NamePersonalisation


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class Ruleset(Item["Ruleset"]):
    definition: Optional[str] = None
    scope: Optional[str] = None
    type: Optional[str] = None


class RulesetScheme(ItemScheme[Ruleset]):
    _Item = Ruleset


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class Transformation(Item["Transformation"]):
    expression: Optional[str] = None
    result: Optional[str] = None


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class UserDefinedOperator(Item["UserDefinedOperator"]):
    definition: Optional[str] = None


class UserDefinedOperatorScheme(ItemScheme[UserDefinedOperator]):
    _Item = UserDefinedOperator


@dataclass
class VTLSpaceKey:
    key: str


class FromVTLSpaceKey(VTLSpaceKey):
    pass


class ToVTLSpaceKey(VTLSpaceKey):
    pass


class VTLMapping(Item["VTLMapping"]):
    pass


#: Mappings from SDMX to VTL.
SDMXtoVTL = Enum("SDMXtoVTL", "basic pivot basic-a2m pivot-a2m")

#: Mappings from VTL to SDMX.
VTLtoSDMX = Enum("VTLtoSDMX", "basic unpivot m2a")


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class VTLConceptMapping(VTLMapping):
    concept_alias: Optional[Concept] = None


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class VTLDataflowMapping(VTLMapping):
    dataflow_alias: Optional[BaseDataflow] = None
    from_vtl_method: Sequence[VTLSpaceKey] = field(default_factory=list)
    from_vtl_superspace: Optional[VTLSpaceKey] = None
    to_vtl_method: Optional[SDMXtoVTL] = None
    to_vtl_subspace: Sequence[VTLSpaceKey] = field(default_factory=list)


class VTLMappingScheme(ItemScheme[VTLMapping]):
    _Item = VTLMapping


@dataclass
class TransformationScheme(ItemScheme[Transformation]):
    _Item = Transformation

    custom_type_scheme: Optional[CustomTypeScheme] = None
    name_personalisation_scheme: Optional[NamePersonalisationScheme] = None
    ruleset_scheme: Optional[RulesetScheme] = None
    user_defined_operator_scheme: Optional[UserDefinedOperatorScheme] = None
    vtl_mapping_scheme: Optional[VTLMappingScheme] = None

    def update_ref(self, ref):
        for f in direct_fields(self.__class__):
            if isinstance(ref, get_args(f.type)[0]):
                setattr(self, f.name, ref)
                return
        raise TypeError(type(ref))


class BaseContentConstraint:
    """ABC for SDMX 2.1 and 3.0 ContentConstraint."""


# Internal

#: The SDMX-IM defines 'packages'; these are used in URNs.
PACKAGE = dict()

_PACKAGE_CLASS: Dict[str, set] = {
    "base": {"Agency", "AgencyScheme", "DataProvider", "DataProviderScheme"},
    "categoryscheme": {"Category", "Categorisation", "CategoryScheme"},
    "codelist": {
        "Code",
        "Codelist",
        "HierarchicalCode",
        "HierarchicalCodelist",  # SDMX 2.1
        "Hierarchy",
        "Level",
    },
    "conceptscheme": {"Concept", "ConceptScheme"},
    "datastructure": {
        "DataflowDefinition",  # SDMX 2.1
        "Dataflow",  # SDMX 3.0
        "DataStructureDefinition",
        "StructureUsage",
    },
    "mapping": {"CodelistMap", "StructureSet"},
    "metadatastructure": {
        "MetadataflowDefinition",  # SDMX 2.1
        "Metadataflow",  # SDMX 3.0
        "MetadataStructureDefinition",
    },
    "registry": {"ContentConstraint", "ProvisionAgreement"},
    "transformation": {
        "CustomTypeScheme",
        "NamePersonalisationScheme",
        "RulesetScheme",
        "UserDefinedOperatorScheme",
    },
}

for package, classes in _PACKAGE_CLASS.items():
    PACKAGE.update({cls: package for cls in classes})

PARENT = {
    Agency: AgencyScheme,
    Category: CategoryScheme,
    Code: Codelist,
    Concept: ConceptScheme,
    Dimension: DimensionDescriptor,
    TimeDimension: DimensionDescriptor,
    DataProvider: DataProviderScheme,
    GroupDimensionDescriptor: BaseDataStructureDefinition,
}


@dataclass
class ClassFinder:
    module_name: str
    name_map: Dict[str, str] = field(default_factory=dict)
    parent_map: Dict[type, type] = field(default_factory=dict)

    def __post_init__(self):
        self._module = sys.modules[self.module_name]
        self._parent = ChainMap(PARENT, self.parent_map)

    @lru_cache()
    def get_class(self, name: Union[str, Resource], package=None) -> Optional[Type]:
        """Return a class for `name` and (optional) `package` names."""
        if isinstance(name, Resource):
            # Convert a Resource enumeration value to a string

            # Expected class name in lower case; maybe just the enumeration value
            match = Resource.class_name(name).lower()

            # Match class names in lower case. If no match or >2, only() returns None,
            # and KeyError occurs below
            name = only(filter(lambda g: g.lower() == match, dir(self._module)))

        # Change names that differ between URNs and full class names
        name = self.name_map.get(name, name)

        try:
            cls = getattr(self._module, name)
        except (AttributeError, TypeError):
            return None

        if package and package != PACKAGE[cls.__name__]:
            raise ValueError(f"Package {repr(package)} invalid for {name}")

        return cls

    def parent_class(self, cls):
        """Return the class that contains objects of type `cls`.

        For example, if `cls` is :class:`.PrimaryMeasure`, returns
        :class:`.v21.MeasureDescriptor`.
        """
        return self._parent[cls]

    def dir(self):
        """For module.__dir__."""
        return sorted(self._module.__all__ + __all__)

    def getattr(self, name):
        """For module.__getattr__."""
        try:
            return globals()[name]
        except KeyError:
            raise AttributeError(name)

    # To allow lru_cache() above
    def __hash__(self):
        return hash(self.module_name)



"""Classes for SDMX messages.

:class:`Message` and related classes are not defined in the SDMX
:doc:`information model <implementation>`, but in the :ref:`SDMX-ML standard <formats>`.

:mod:`sdmx` also uses :class:`DataMessage` to encapsulate SDMX-JSON data returned by
data sources.
"""
import logging
from dataclasses import dataclass, field, fields
from datetime import datetime
from operator import attrgetter
from typing import TYPE_CHECKING, List, Optional, Text, Type, Union, get_args

from sdmx import model
from sdmx.dictlike import DictLike, DictLikeDescriptor, summarize_dictlike
from sdmx.format import Version
from sdmx.model import common, v21, v30
from sdmx.model.internationalstring import (
    InternationalString,
    InternationalStringDescriptor,
)
from sdmx.util import compare, direct_fields

if TYPE_CHECKING:
    import requests

log = logging.getLogger(__name__)


def _summarize(obj, include: Optional[List[str]] = None):
    """Helper method for __repr__ on Header and Message (sub)classes."""
    import requests

    include = include or list(map(attrgetter("name"), fields(obj)))
    for name in include:
        attr = getattr(obj, name)
        if attr is None:
            continue
        elif isinstance(attr, datetime):
            s_attr = repr(attr.isoformat())
        elif isinstance(attr, requests.Response):
            # Don't use repr(), which displays the entire response body
            s_attr = str(attr)
        else:
            s_attr = repr(attr)

        yield f"{name}: {s_attr}"


@dataclass
class Header:
    """Header of an SDMX-ML message.

    SDMX-JSON messages do not have headers.
    """

    #: (optional) Error code for the message.
    error: Optional[Text] = None
    #: Date and time at which the data was extracted.
    extracted: Optional[datetime] = None
    #: Identifier for the message.
    id: Optional[Text] = None
    #: Date and time at which the message was generated.
    prepared: Optional[datetime] = None
    #: Start of the time period covered by a :class:`.DataMessage`.
    reporting_begin: Optional[datetime] = None
    #: End of the time period covered by a :class:`.DataMessage`.
    reporting_end: Optional[datetime] = None
    #: Intended recipient of the message, e.g. the user's name for an
    #: authenticated service.
    receiver: Optional[model.Agency] = None
    #: The :class:`.Agency` associated with the data :class:`~.source.Source`.
    sender: Optional[model.Agency] = None
    #:
    source: InternationalStringDescriptor = InternationalStringDescriptor()
    #:
    test: bool = False

    def __repr__(self):
        """String representation."""
        lines = ["<Header>"]
        lines.extend(_summarize(self))
        return "\n  ".join(lines)

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two Headers are the same if their corresponding attributes are equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return all(
            compare(f, self, other, strict)
            for f in map(attrgetter("name"), fields(self))
        )


@dataclass
class Footer:
    """Footer of an SDMX-ML message.

    SDMX-JSON messages do not have footers.
    """

    #:
    severity: Optional[str] = None
    #: The body text of the Footer contains zero or more blocks of text.
    text: List[model.InternationalString] = field(default_factory=list)
    #:
    code: Optional[int] = None

    def __post_init__(self):
        # Convert non-IS members to IS
        self.text = [
            t if isinstance(t, InternationalString) else InternationalString(t)
            for t in self.text
        ]

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two Footers are the same if their :attr:`code`, :attr:`severity`, and
        :attr:`text` are equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return all(
            compare(f, self, other, strict)
            for f in map(attrgetter("name"), fields(self))
        )


@dataclass
class Message:
    #: SDMX version.
    version: Version = Version["2.1"]

    #: :class:`Header` instance.
    header: Header = field(default_factory=Header)
    #: (optional) :class:`Footer` instance.
    footer: Optional[Footer] = None
    #: :class:`requests.Response` instance for the response to the HTTP request that
    #: returned the Message. This is not part of the SDMX standard.
    response: Optional["requests.Response"] = None

    def __str__(self):
        return repr(self)

    def __repr__(self):
        """String representation."""
        lines = [
            f"<sdmx.{self.__class__.__name__}>",
            repr(self.header).replace("\n", "\n  "),
        ]
        lines.extend(_summarize(self, ["footer", "response"]))
        return "\n  ".join(lines)

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two Messages are the same if their :attr:`header` and :attr:`footer` compare
        equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return self.header.compare(other.header, strict) and (
            self.footer is other.footer is None
            or self.footer.compare(other.footer, strict)
        )


class ErrorMessage(Message):
    pass


@dataclass
class StructureMessage(Message):
    #: Collection of :class:`.Categorisation`.
    categorisation: DictLikeDescriptor[str, model.Categorisation] = DictLikeDescriptor()
    #: Collection of :class:`.CategoryScheme`.
    category_scheme: DictLikeDescriptor[
        str, model.CategoryScheme
    ] = DictLikeDescriptor()
    #: Collection of :class:`.Codelist`.
    codelist: DictLikeDescriptor[str, model.Codelist] = DictLikeDescriptor()
    #: Collection of :class:`.HierarchicalCodelist`.
    hierarchical_codelist: DictLikeDescriptor[
        str, v21.HierarchicalCodelist
    ] = DictLikeDescriptor()
    #: Collection of :class:`.v30.Hierarchy`.
    hierarchy: DictLikeDescriptor[str, v30.Hierarchy] = DictLikeDescriptor()
    #: Collection of :class:`.ConceptScheme`.
    concept_scheme: DictLikeDescriptor[str, model.ConceptScheme] = DictLikeDescriptor()
    #: Collection of :class:`.ContentConstraint`.
    constraint: DictLikeDescriptor[str, model.BaseConstraint] = DictLikeDescriptor()
    #: Collection of :class:`Dataflow(Definition) <.BaseDataflow>`.
    dataflow: DictLikeDescriptor[str, model.BaseDataflow] = DictLikeDescriptor()
    #: Collection of :class:`MetadataStructureDefinition
    #: <.BaseMetadataStructureDefinition>`.
    metadatastructure: DictLikeDescriptor[
        str, model.BaseMetadataStructureDefinition
    ] = DictLikeDescriptor()
    #: Collection of :class:`Metadataflow(Definition) <.BaseMetadataflow>`.
    metadataflow: DictLikeDescriptor[str, model.BaseMetadataflow] = DictLikeDescriptor()
    #: Collection of :class:`DataStructureDefinition <.BaseDataStructureDefinition>`.
    structure: DictLikeDescriptor[
        str, model.BaseDataStructureDefinition
    ] = DictLikeDescriptor()
    #: Collection of :class:`.StructureSet`.
    structureset: DictLikeDescriptor[str, v21.StructureSet] = DictLikeDescriptor()
    #: Collection of :class:`.OrganisationScheme`.
    organisation_scheme: DictLikeDescriptor[
        str, model.OrganisationScheme
    ] = DictLikeDescriptor()
    #: Collection of :class:`.ProvisionAgreement`.
    provisionagreement: DictLikeDescriptor[
        str, model.ProvisionAgreement
    ] = DictLikeDescriptor()

    #: Collection of :class:`.CustomTypeScheme`.
    custom_type_scheme: DictLikeDescriptor[
        str, model.CustomTypeScheme
    ] = DictLikeDescriptor()
    #: Collection of :class:`.NamePersonalisationScheme`.
    name_personalisation_scheme: DictLikeDescriptor[
        str, model.NamePersonalisationScheme
    ] = DictLikeDescriptor()
    #: Collection of :class:`.RulesetScheme`.
    ruleset_scheme: DictLikeDescriptor[str, model.RulesetScheme] = DictLikeDescriptor()
    #: Collection of :class:`.VTLMappingScheme`.
    vtl_mapping_scheme: DictLikeDescriptor[
        str, model.VTLMappingScheme
    ] = DictLikeDescriptor()
    #: Collection of :class:`.TransformationScheme`.
    transformation_scheme: DictLikeDescriptor[
        str, model.TransformationScheme
    ] = DictLikeDescriptor()
    #: Collection of :class:`.UserDefinedOperatorScheme`.
    user_defined_operator_scheme: DictLikeDescriptor[
        str, model.UserDefinedOperatorScheme
    ] = DictLikeDescriptor()

    #: Collection of :class:`.ValueList` (SDMX 3.0 only).
    valuelist: DictLikeDescriptor[str, v30.ValueList] = DictLikeDescriptor()

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two StructureMessages compare equal if :meth:`.DictLike.compare` is :obj:`True`
        for each of the object collection attributes.

        Parameters
        ----------
        strict : bool, optional
            Passed to :meth:`.DictLike.compare`.
        """
        return super().compare(other, strict) and all(
            getattr(self, f.name).compare(getattr(other, f.name), strict)
            for f in direct_fields(self.__class__)
        )

    def add(self, obj: model.IdentifiableArtefact):
        """Add `obj` to the StructureMessage."""
        for f in direct_fields(self.__class__):
            # NB for some reason mypy complains here, but not in __contains__(), below
            if isinstance(obj, get_args(f.type)[1]):
                getattr(self, f.name)[obj.id] = obj
                return
        raise TypeError(type(obj))

    def get(
        self, obj_or_id: Union[str, model.IdentifiableArtefact]
    ) -> Optional[model.IdentifiableArtefact]:
        """Retrieve `obj_or_id` from the StructureMessage.

        Parameters
        ----------
        obj_or_id : str or .IdentifiableArtefact
            If an IdentifiableArtefact, return an object of the same class and
            :attr:`~.IdentifiableArtefact.id`; if :class:`str`, an object with this ID.

        Returns
        -------
        .IdentifiableArtefact
            with the given ID and possibly class.
        None
            if there is no match.

        Raises
        ------
        ValueError
            if `obj_or_id` is a string and there are ≥2 objects (of different classes)
            with the same ID.
        """
        id = (
            obj_or_id.id
            if isinstance(obj_or_id, model.IdentifiableArtefact)
            else obj_or_id
        )

        candidates: List[model.IdentifiableArtefact] = list(
            filter(
                None,
                map(
                    lambda f: getattr(self, f.name).get(id),
                    direct_fields(self.__class__),
                ),
            )
        )

        if len(candidates) > 1:
            raise ValueError(f"ambiguous; {repr(obj_or_id)} matches {repr(candidates)}")

        return candidates[0] if len(candidates) == 1 else None

    def iter_collections(self):
        """Iterate over collections."""
        for f in direct_fields(self.__class__):
            yield f.name, get_args(f.type)[1]

    def objects(self, cls):
        """Get a reference to the attribute for objects of type `cls`.

        For example, if `cls` is the class :class:`DataStructureDefinition` (not an
        instance), return a reference to :attr:`structure`.
        """
        for f in direct_fields(self.__class__):
            if issubclass(cls, get_args(f.type)[1]):
                return getattr(self, f.name)
        raise TypeError(cls)

    def __contains__(self, item):
        """Return :obj:`True` if `item` is in the StructureMessage."""
        for f in direct_fields(self.__class__):
            if isinstance(item, get_args(f.type)[1]):
                return item in getattr(self, f.name).values()
        raise TypeError(f"StructureMessage has no collection of {type(item)}")

    def __repr__(self):
        """String representation."""
        lines = [super().__repr__()]

        # StructureMessage contents
        for attr in self.__dict__.values():
            if isinstance(attr, DictLike) and attr:
                lines.append(summarize_dictlike(attr))

        return "\n  ".join(lines)


@dataclass
class DataMessage(Message):
    """SDMX Data Message.

    .. note:: A DataMessage may contain zero or more :class:`.DataSet`, so
       :attr:`data` is a list. To retrieve the first (and possibly only)
       data set in the message, access the first element of the list:
       ``msg.data[0]``.
    """

    #: :class:`list` of :class:`.DataSet`.
    data: List[model.BaseDataSet] = field(default_factory=list)
    #: :class:`.DataflowDefinition` that contains the data.
    dataflow: Optional[model.BaseDataflow] = None
    #: The "dimension at observation level".
    observation_dimension: Optional[
        Union[
            model._AllDimensions,
            model.DimensionComponent,
            List[model.DimensionComponent],
        ]
    ] = None

    def __post_init__(self):
        if self.dataflow is None:
            # Create a default of the appropriate class
            self.dataflow = {
                Version["2.1"]: v21.DataflowDefinition,
                Version["3.0.0"]: v30.Dataflow,
            }[self.version]()

    # Convenience access
    @property
    def structure(self):
        """DataStructureDefinition used in the :attr:`dataflow`."""
        return self.dataflow.structure

    @property
    def structure_type(self) -> Type[common.Structure]:
        """:class:`.Structure` subtype describing the contained (meta)data."""
        return {
            Version["2.1"]: v21.DataStructureDefinition,
            Version["3.0.0"]: v30.DataStructureDefinition,
        }[self.version]

    def __repr__(self):
        """String representation."""
        lines = [super().__repr__()]

        # DataMessage contents
        if self.data:
            lines.append("DataSet ({})".format(len(self.data)))
        lines.extend(_summarize(self, ("dataflow", "observation_dimension")))

        return "\n  ".join(lines)

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two DataMessages are the same if:

        - :meth:`.Message.compare` is :obj:`True`
        - their :attr:`dataflow` and :attr:`observation_dimension` compare equal.
        - they have the same number of :class:`DataSets <DataSet>`, and
        - corresponding DataSets compare equal (see :meth:`.DataSet.compare`).

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`.compare`.
        """
        return (
            super().compare(other, strict)
            and compare("dataflow", self, other, strict)
            and compare("observation_dimension", self, other, strict)
            and len(self.data) == len(other.data)
            and all(ds[0].compare(ds[1], strict) for ds in zip(self.data, other.data))
        )


@dataclass
class MetadataMessage(DataMessage):
    """SDMX Metadata Message."""

    @property
    def structure_type(self) -> Type[common.Structure]:
        return {
            Version["2.1"]: v21.MetadataStructureDefinition,
            Version["3.0.0"]: v30.MetadataStructureDefinition,
        }[self.version]



import logging
from collections.abc import Mapping

import pandas as pd
import pytest

import sdmx
from sdmx import message
from sdmx.model import v21 as model
from sdmx.model.v21 import Key
from sdmx.testing import MessageTest


class StructuredMessageTest(MessageTest):
    """Variant of MessageTest for structure-specific messages."""

    directory = "ECB_EXR"
    dsd_filename: str

    # Fixtures
    @pytest.fixture(scope="class")
    def dsd(self, path):
        yield sdmx.read_sdmx(path / self.dsd_filename).structure[0]

    @pytest.fixture(scope="class")
    def msg(self, path, dsd):
        yield sdmx.read_sdmx(path / self.filename, dsd=dsd)

    # Tests for every class
    def test_msg(self, caplog, path, dsd):
        # The message can be parsed
        with caplog.at_level(logging.WARNING):
            sdmx.read_sdmx(path / self.filename, dsd=dsd)

        # No warnings are emitted per missing dsd= argument (khaeru/sdmx#146)
        assert [] == caplog.messages

    def test_structured_by(self, dsd, msg):
        # The DSD was used to parse the message
        assert msg.data[0].structured_by is dsd


class TestFlatDataSet(StructuredMessageTest):
    filename = "ng-flat-ss.xml"
    dsd_filename = "ng-structure-full.xml"

    def test_msg_type(self, msg):
        assert isinstance(msg, message.DataMessage)
        # assert msg.data[0].dim_at_obs == 'AllDimensions'

    def test_dataset_cls(self, msg):
        assert isinstance(msg.data[0], model.DataSet)

    def test_generic_obs(self, msg):
        data = msg.data[0]
        # empty series list
        assert len(list(data.series)) == 0
        obs_list = data.obs
        assert len(obs_list) == 12
        o0 = obs_list[0]

        # Flat data set → all six dimensions are at the observation level
        assert len(o0) == 6
        assert o0.key.FREQ == "M"
        assert o0.key.CURRENCY == "CHF"
        assert o0.value == "1.3413"

        # All attributes are at the observation level
        assert len(o0.attrib) == 7
        assert o0.attrib.OBS_STATUS == "A"
        assert o0.attrib.DECIMALS == "4"

    def test_write2pandas(self, msg):
        data_series = sdmx.to_pandas(msg, attributes="")
        assert isinstance(data_series, pd.Series)


class TestSeriesDataSet(StructuredMessageTest):
    filename = "ng-ts-gf-ss.xml"
    dsd_filename = "ng-structure-full.xml"

    def test_msg_type(self, dsd, msg):
        # assert msg.data[0].dim_at_obs == 'TIME_PERIOD'
        pass

    def test_dataset_cls(self, msg):
        assert isinstance(msg.data[0], model.DataSet)

    def test_obs(self, msg):
        data = msg.data[0]
        assert 12 == len(data.obs)

        series_list = list(data.series)
        assert 4 == len(series_list)

        s3 = series_list[3]
        assert isinstance(s3, model.SeriesKey)

        # Time series data set → five dimensions are at the SeriesKey level
        assert 5 == len(s3)
        assert "USD" == s3.CURRENCY

        # 5 of 7 attributes are at the Observation level
        assert 5 == len(s3.attrib)
        assert "4" == s3.attrib.DECIMALS

        obs_list = data.series[s3]
        assert 3 == len(obs_list)
        o2 = obs_list[2]

        # One remaining dimension is at the Observation Level
        assert 1 == len(o2.dimension)
        assert Key(TIME_PERIOD="2010-10") == o2.dim
        assert "1.3898" == o2.value

        # Two remaining attributes are at the Observation level
        assert 2 == len(o2.attached_attribute)
        assert "A" == o2.attrib.OBS_STATUS

    def test_pandas(self, msg):
        data = msg.data[0]

        # Expected number of observations and series
        assert len(data.obs) == 12
        assert len(data.series) == 4

        # Single series can be converted to pandas
        s3 = sdmx.to_pandas(data.series[3], attributes="")
        assert isinstance(s3, pd.Series)
        # With expected values
        assert s3.iloc[0] == 1.2894

        # Single series can be converted with attributes
        s3_attr = sdmx.to_pandas(data.series[3], attributes="osgd")

        # yields a DataFrame
        assert isinstance(s3_attr, pd.DataFrame)
        assert s3_attr.shape == (3, 8)

        assert s3_attr.iloc[0].value == 1.2894

        # Attributes of observations can be accessed
        assert s3_attr.iloc[0].OBS_STATUS == "A"

    def test_write2pandas(self, msg):
        df = sdmx.to_pandas(msg, attributes="")
        assert isinstance(df, pd.Series)
        assert df.shape == (12,)
        # with metadata
        df = sdmx.to_pandas(msg, attributes="osgd")
        assert df.shape == (12, 8)
        assert df.iloc[1].OBS_STATUS == "A"


class TestSeriesDataSet2(StructuredMessageTest):
    filename = "ng-ts-ss.xml"
    dsd_filename = "ng-structure-full.xml"

    def test_msg_type(self, dsd, msg):
        # assert msg.data[0].dim_at_obs == 'TIME_PERIOD'
        pass

    def test_dataset_cls(self, msg):
        assert isinstance(msg.data[0], model.DataSet)

    def test_structured_obs(self, msg):
        data = msg.data[0]

        # Expected number of observations and series
        assert 12 == len(data.obs)
        assert 4 == len(data.series)

        # SeriesKey is accessible by index using DictLike
        s3 = list(data.series.keys())[3]
        assert isinstance(s3, model.SeriesKey)

        # SeriesKey has expected number of dimensions and values
        assert 5 == len(s3)
        assert "USD" == s3.CURRENCY

        # SeriesKey has expected number of attributes and values
        assert 5 == len(s3.attrib)
        assert "4" == s3.attrib.DECIMALS

        # Series observations can be accessed
        obs_list = data.series[s3]
        assert 3 == len(obs_list)
        o2 = obs_list[2]
        assert 1 == len(o2.dimension)
        assert Key(TIME_PERIOD="2010-10") == o2.dim
        assert "1.3898" == o2.value
        assert "A" == o2.attrib.OBS_STATUS

    def test_dataframe(self, msg):
        data = msg.data[0]
        s = sdmx.to_pandas(data, attributes="")
        assert isinstance(s, pd.Series)
        assert len(s) == 12


class TestSeriesData_SiblingGroup_TS(StructuredMessageTest):
    filename = "sg-ts-ss.xml"
    dsd_filename = "sg-structure.xml"

    def test_groups(self, msg):
        data = msg.data[0]
        assert len(data.group) == 4
        assert len(data.series) == 4
        g2 = list(data.group.keys())[2]
        assert g2.CURRENCY == "JPY"
        assert g2.attrib.TITLE == "ECB reference exchange rate, Japanese yen/Euro"

        # Check direct attributes of SeriesKey
        s = list(data.series)[0]
        assert 4 == len(s.attrib)

        # Check group attributes of a series
        g_attrib = s.group_attrib
        assert isinstance(g_attrib, Mapping)
        assert 1 == len(g_attrib)


class TestSeriesData_RateGroup_TS(StructuredMessageTest):
    filename = "rg-ts-ss.xml"
    dsd_filename = "rg-structure.xml"

    def test_dsd(self, dsd):
        ad = dsd.attributes
        for id in ("CONF_STATUS_OBS", "OBS_STATUS"):
            assert isinstance(ad.get(id).related_to, model.PrimaryMeasureRelationship)
        for id in ("COLL_METHOD", "TITLE"):
            assert isinstance(ad.get(id).related_to, model.GroupRelationship)
        for id in ("DECIMALS", "UNIT_MEASURE", "UNIT_MULT"):
            assert isinstance(ad.get(id).related_to, model.DimensionRelationship)

    def test_groups(self, msg):
        data = msg.data[0]
        assert len(data.group) == 5
        assert len(data.series) == 4
        g2 = list(data.group.keys())[2]
        assert g2.CURRENCY == "GBP"
        assert (
            g2.attrib.TITLE == "ECB reference exchange rate, U.K. Pound sterling /Euro"
        )

        # Check group attributes of a series
        s = list(data.series)[0]
        g_attrib = s.group_attrib
        assert isinstance(g_attrib, Mapping)
        assert 5 == len(g_attrib)
